#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#pylint: disable=
"""
File       : general.py
Author     : Rolf Verberg <rolfverberg AT gmail dot com>
Description: A collection of general modules
"""
# RV write function that returns a list of peak indices for a given plot
# RV use raise_error concept on more functions

# System modules
from ast import literal_eval
from logging import getLogger
from os import path as os_path
from os import (
    access,
    R_OK,
)
from re import compile as re_compile
from re import split as re_split
from re import sub as re_sub
from sys import float_info

# Third party modules
import numpy as np
try:
    import matplotlib.pyplot as plt
    import matplotlib.lines as mlines
    from matplotlib.widgets import Button
except ImportError:
    pass

logger = getLogger(__name__)


def depth_list(_list):
    """Return the depth of a list."""
    return isinstance(_list, list) and 1+max(map(depth_list, _list))


def depth_tuple(_tuple):
    """Return the depth of a tuple."""
    return isinstance(_tuple, tuple) and 1+max(map(depth_tuple, _tuple))


def unwrap_tuple(_tuple):
    """Unwrap a tuple."""
    if depth_tuple(_tuple) > 1 and len(_tuple) == 1:
        _tuple = unwrap_tuple(*_tuple)
    return _tuple


def illegal_value(value, name, location=None, raise_error=False, log=True):
    """Print illegal value message and/or raise error."""
    if not isinstance(location, str):
        location = ''
    else:
        location = f'in {location} '
    if isinstance(name, str):
        error_msg = \
            f'Illegal value for {name} {location}({value}, {type(value)})'
    else:
        error_msg = f'Illegal value {location}({value}, {type(value)})'
    if log:
        logger.error(error_msg)
    if raise_error:
        raise ValueError(error_msg)


def illegal_combination(
        value1, name1, value2, name2, location=None, raise_error=False,
        log=True):
    """Print illegal combination message and/or raise error."""
    if not isinstance(location, str):
        location = ''
    else:
        location = f'in {location} '
    if isinstance(name1, str):
        error_msg = f'Illegal combination for {name1} and {name2} {location}' \
            f'({value1}, {type(value1)} and {value2}, {type(value2)})'
    else:
        error_msg = f'Illegal combination {location}' \
            f'({value1}, {type(value1)} and {value2}, {type(value2)})'
    if log:
        logger.error(error_msg)
    if raise_error:
        raise ValueError(error_msg)


def test_ge_gt_le_lt(
        ge, gt, le, lt, func, location=None, raise_error=False, log=True):
    """
    Check individual and mutual validity of ge, gt, le, lt qualifiers.

    :param func: Test for integers or numbers
    :type func: callable: is_int, is_num
    :return: True upon success or False when mutually exlusive
    :rtype: bool
    """
    if ge is None and gt is None and le is None and lt is None:
        return True
    if ge is not None:
        if not func(ge):
            illegal_value(ge, 'ge', location, raise_error, log)
            return False
        if gt is not None:
            illegal_combination(ge, 'ge', gt, 'gt', location, raise_error, log)
            return False
    elif gt is not None and not func(gt):
        illegal_value(gt, 'gt', location, raise_error, log)
        return False
    if le is not None:
        if not func(le):
            illegal_value(le, 'le', location, raise_error, log)
            return False
        if lt is not None:
            illegal_combination(le, 'le', lt, 'lt', location, raise_error, log)
            return False
    elif lt is not None and not func(lt):
        illegal_value(lt, 'lt', location, raise_error, log)
        return False
    if ge is not None:
        if le is not None and ge > le:
            illegal_combination(ge, 'ge', le, 'le', location, raise_error, log)
            return False
        if lt is not None and ge >= lt:
            illegal_combination(ge, 'ge', lt, 'lt', location, raise_error, log)
            return False
    elif gt is not None:
        if le is not None and gt >= le:
            illegal_combination(gt, 'gt', le, 'le', location, raise_error, log)
            return False
        if lt is not None and gt >= lt:
            illegal_combination(gt, 'gt', lt, 'lt', location, raise_error, log)
            return False
    return True


def range_string_ge_gt_le_lt(ge=None, gt=None, le=None, lt=None):
    """
    Return a range string representation matching the ge, gt, le, lt
    qualifiers. Does not validate the inputs, do that as needed before
    calling.
    """
    range_string = ''
    if ge is not None:
        if le is None and lt is None:
            range_string += f'>= {ge}'
        else:
            range_string += f'[{ge}, '
    elif gt is not None:
        if le is None and lt is None:
            range_string += f'> {gt}'
        else:
            range_string += f'({gt}, '
    if le is not None:
        if ge is None and gt is None:
            range_string += f'<= {le}'
        else:
            range_string += f'{le}]'
    elif lt is not None:
        if ge is None and gt is None:
            range_string += f'< {lt}'
        else:
            range_string += f'{lt})'
    return range_string


def is_int(v, ge=None, gt=None, le=None, lt=None, raise_error=False, log=True):
    """
    Value is an integer in range ge <= v <= le or gt < v < lt or some
    combination.

    :return: True if yes or False is no
    :rtype: bool
    """
    return _is_int_or_num(v, 'int', ge, gt, le, lt, raise_error, log)


def is_num(v, ge=None, gt=None, le=None, lt=None, raise_error=False, log=True):
    """
    Value is a number in range ge <= v <= le or gt < v < lt or some
    combination.

    :return: True if yes or False is no
    :rtype: bool
    """
    return _is_int_or_num(v, 'num', ge, gt, le, lt, raise_error, log)


def _is_int_or_num(
        v, type_str, ge=None, gt=None, le=None, lt=None, raise_error=False,
        log=True):
    if type_str == 'int':
        if not isinstance(v, int):
            illegal_value(v, 'v', '_is_int_or_num', raise_error, log)
            return False
        if not test_ge_gt_le_lt(
                ge, gt, le, lt, is_int, '_is_int_or_num', raise_error, log):
            return False
    elif type_str == 'num':
        if not isinstance(v, (int, float)):
            illegal_value(v, 'v', '_is_int_or_num', raise_error, log)
            return False
        if not test_ge_gt_le_lt(
                ge, gt, le, lt, is_num, '_is_int_or_num', raise_error, log):
            return False
    else:
        illegal_value(type_str, 'type_str', '_is_int_or_num', raise_error, log)
        return False
    if ge is None and gt is None and le is None and lt is None:
        return True
    error = False
    if ge is not None and v < ge:
        error = True
        error_msg = f'Value {v} out of range: {v} !>= {ge}'
    if not error and gt is not None and v <= gt:
        error = True
        error_msg = f'Value {v} out of range: {v} !> {gt}'
    if not error and le is not None and v > le:
        error = True
        error_msg = f'Value {v} out of range: {v} !<= {le}'
    if not error and lt is not None and v >= lt:
        error = True
        error_msg = f'Value {v} out of range: {v} !< {lt}'
    if error:
        if log:
            logger.error(error_msg)
        if raise_error:
            raise ValueError(error_msg)
        return False
    return True


def is_int_pair(
        v, ge=None, gt=None, le=None, lt=None, raise_error=False, log=True):
    """
    Value is an integer pair, each in range ge <= v[i] <= le or
    gt < v[i] < lt or ge[i] <= v[i] <= le[i] or gt[i] < v[i] < lt[i]
    or some combination.

    :return: True if yes or False is no
    :rtype: bool
    """
    return _is_int_or_num_pair(v, 'int', ge, gt, le, lt, raise_error, log)


def is_num_pair(
        v, ge=None, gt=None, le=None, lt=None, raise_error=False, log=True):
    """
    Value is a number pair, each in range ge <= v[i] <= le or
    gt < v[i] < lt or ge[i] <= v[i] <= le[i] or gt[i] < v[i] < lt[i]
    or some combination.

    :return: True if yes or False is no
    :rtype: bool
    """
    return _is_int_or_num_pair(v, 'num', ge, gt, le, lt, raise_error, log)


def _is_int_or_num_pair(
        v, type_str, ge=None, gt=None, le=None, lt=None, raise_error=False,
        log=True):
    if type_str == 'int':
        if not (isinstance(v, (tuple, list)) and len(v) == 2
                and isinstance(v[0], int) and isinstance(v[1], int)):
            illegal_value(v, 'v', '_is_int_or_num_pair', raise_error, log)
            return False
        func = is_int
    elif type_str == 'num':
        if not (isinstance(v, (tuple, list)) and len(v) == 2
                and isinstance(v[0], (int, float))
                and isinstance(v[1], (int, float))):
            illegal_value(v, 'v', '_is_int_or_num_pair', raise_error, log)
            return False
        func = is_num
    else:
        illegal_value(
            type_str, 'type_str', '_is_int_or_num_pair', raise_error, log)
        return False
    if ge is None and gt is None and le is None and lt is None:
        return True
    if ge is None or func(ge, log=True):
        ge = 2*[ge]
    elif not _is_int_or_num_pair(
            ge, type_str, raise_error=raise_error, log=log):
        return False
    if gt is None or func(gt, log=True):
        gt = 2*[gt]
    elif not _is_int_or_num_pair(
            gt, type_str, raise_error=raise_error, log=log):
        return False
    if le is None or func(le, log=True):
        le = 2*[le]
    elif not _is_int_or_num_pair(
            le, type_str, raise_error=raise_error, log=log):
        return False
    if lt is None or func(lt, log=True):
        lt = 2*[lt]
    elif not _is_int_or_num_pair(
            lt, type_str, raise_error=raise_error, log=log):
        return False
    if (not func(v[0], ge[0], gt[0], le[0], lt[0], raise_error, log)
            or not func(v[1], ge[1], gt[1], le[1], lt[1], raise_error, log)):
        return False
    return True


def is_int_series(
        t_or_l, ge=None, gt=None, le=None, lt=None, raise_error=False,
        log=True):
    """
    Value is a tuple or list of integers, each in range
    ge <= l[i] <= le or gt < l[i] < lt or some combination.
    """
    if not test_ge_gt_le_lt(
            ge, gt, le, lt, is_int, 'is_int_series', raise_error, log):
        return False
    if not isinstance(t_or_l, (tuple, list)):
        illegal_value(t_or_l, 't_or_l', 'is_int_series', raise_error, log)
        return False
    if any(not is_int(v, ge, gt, le, lt, raise_error, log) for v in t_or_l):
        return False
    return True


def is_num_series(
        t_or_l, ge=None, gt=None, le=None, lt=None, raise_error=False,
        log=True):
    """
    Value is a tuple or list of numbers, each in range ge <= l[i] <= le
    or gt < l[i] < lt or some combination.
    """
    if not test_ge_gt_le_lt(
            ge, gt, le, lt, is_int, 'is_int_series', raise_error, log):
        return False
    if not isinstance(t_or_l, (tuple, list)):
        illegal_value(t_or_l, 't_or_l', 'is_num_series', raise_error, log)
        return False
    if any(not is_num(v, ge, gt, le, lt, raise_error, log) for v in t_or_l):
        return False
    return True


def is_str_series(t_or_l, raise_error=False, log=True):
    """
    Value is a tuple or list of strings.
    """
    if (not isinstance(t_or_l, (tuple, list))
            or any(not isinstance(s, str) for s in t_or_l)):
        illegal_value(t_or_l, 't_or_l', 'is_str_series', raise_error, log)
        return False
    return True


def is_dict_series(t_or_l, raise_error=False, log=True):
    """
    Value is a tuple or list of dictionaries.
    """
    if (not isinstance(t_or_l, (tuple, list))
            or any(not isinstance(d, dict) for d in t_or_l)):
        illegal_value(t_or_l, 't_or_l', 'is_dict_series', raise_error, log)
        return False
    return True


def is_dict_nums(d, raise_error=False, log=True):
    """
    Value is a dictionary with single number values
    """
    if (not isinstance(d, dict)
            or any(not is_num(v, log=False) for v in d.values())):
        illegal_value(d, 'd', 'is_dict_nums', raise_error, log)
        return False
    return True


def is_dict_strings(d, raise_error=False, log=True):
    """
    Value is a dictionary with single string values
    """
    if (not isinstance(d, dict)
            or any(not isinstance(v, str) for v in d.values())):
        illegal_value(d, 'd', 'is_dict_strings', raise_error, log)
        return False
    return True


def is_index(v, ge=0, lt=None, raise_error=False, log=True):
    """
    Value is an array index in range ge <= v < lt. NOTE lt IS NOT
    included!
    """
    if isinstance(lt, int):
        if lt <= ge:
            illegal_combination(
                ge, 'ge', lt, 'lt', 'is_index', raise_error, log)
            return False
    return is_int(v, ge=ge, lt=lt, raise_error=raise_error, log=log)


def is_index_range(v, ge=0, le=None, lt=None, raise_error=False, log=True):
    """
    Value is an array index range in range ge <= v[0] <= v[1] <= le or
    ge <= v[0] <= v[1] < lt. NOTE le IS included!
    """
    if not is_int_pair(v, raise_error=raise_error, log=log):
        return False
    if not test_ge_gt_le_lt(
            ge, None, le, lt, is_int, 'is_index_range', raise_error, log):
        return False
    if (not ge <= v[0] <= v[1] or (le is not None and v[1] > le)
            or (lt is not None and v[1] >= lt)):
        if le is not None:
            error_msg = \
                f'Value {v} out of range: !({ge} <= {v[0]} <= {v[1]} <= {le})'
        else:
            error_msg = \
                f'Value {v} out of range: !({ge} <= {v[0]} <= {v[1]} < {lt})'
        if log:
            logger.error(error_msg)
        if raise_error:
            raise ValueError(error_msg)
        return False
    return True


def index_nearest(a, value):
    """Return index of nearest array value."""
    a = np.asarray(a)
    if a.ndim > 1:
        raise ValueError(
            f'Invalid array dimension for parameter a ({a.ndim}, {a})')
    # Round up for .5
    value *= 1.0+float_info.epsilon
    return (int)(np.argmin(np.abs(a-value)))


def index_nearest_low(a, value):
    """Return index of nearest array value, rounded down"""
    a = np.asarray(a)
    if a.ndim > 1:
        raise ValueError(
            f'Invalid array dimension for parameter a ({a.ndim}, {a})')
    index = int(np.argmin(np.abs(a-value)))
    if value < a[index] and index > 0:
        index -= 1
    return index


def index_nearest_upp(a, value):
    """Return index of nearest array value, rounded upp."""
    a = np.asarray(a)
    if a.ndim > 1:
        raise ValueError(
            f'Invalid array dimension for parameter a ({a.ndim}, {a})')
    index = int(np.argmin(np.abs(a-value)))
    if value > a[index] and index < a.size-1:
        index += 1
    return index


def round_to_n(x, n=1):
    """Round to a specific number of decimals."""
    if x == 0.0:
        return 0
    return type(x)(round(x, n-1-int(np.floor(np.log10(abs(x))))))


def round_up_to_n(x, n=1):
    """Round up to a specific number of decimals."""
    x_round = round_to_n(x, n)
    if abs(x/x_round) > 1.0:
        x_round += np.sign(x) * 10**(np.floor(np.log10(abs(x)))+1-n)
    return type(x)(x_round)


def trunc_to_n(x, n=1):
    """Truncate to a specific number of decimals."""
    x_round = round_to_n(x, n)
    if abs(x_round/x) > 1.0:
        x_round -= np.sign(x) * 10**(np.floor(np.log10(abs(x)))+1-n)
    return type(x)(x_round)


def almost_equal(a, b, sig_figs):
    """
    Check if equal to within a certain number of significant digits.
    """
    if is_num(a) and is_num(b):
        return abs(round_to_n(a-b, sig_figs)) < pow(10, 1-sig_figs)
    raise ValueError(
        f'Invalid value for a or b in almost_equal (a: {a}, {type(a)}, '
        f'b: {b}, {type(b)})')


def string_to_list(s, split_on_dash=True, remove_duplicates=True, sort=True):
    """
    Return a list of numbers by splitting/expanding a string on any
    combination of commas, whitespaces, or dashes (when
    split_on_dash=True).
    e.g: '1, 3, 5-8, 12 ' -> [1, 3, 5, 6, 7, 8, 12]
    """
    if not isinstance(s, str):
        illegal_value(s, 's', location='string_to_list')
        return None
    if not s:
        return []
    try:
        list1 = re_split(r'\s+,\s+|\s+,|,\s+|\s+|,', s.strip())
    except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError):
        return None
    if split_on_dash:
        try:
            l_of_i = []
            for v in list1:
                list2 = [
                    literal_eval(x)
                    for x in re_split(r'\s+-\s+|\s+-|-\s+|\s+|-', v)]
                if len(list2) == 1:
                    l_of_i += list2
                elif len(list2) == 2 and list2[1] > list2[0]:
                    l_of_i += list(range(list2[0], 1+list2[1]))
                else:
                    raise ValueError
        except (ValueError, TypeError, SyntaxError, MemoryError,
                RecursionError):
            return None
    else:
        l_of_i = [literal_eval(x) for x in list1]
    if remove_duplicates:
        l_of_i = list(dict.fromkeys(l_of_i))
    if sort:
        l_of_i = sorted(l_of_i)
    return l_of_i


def get_trailing_int(string):
    """Get the trailing integer in a string."""
    index_regex = re_compile(r'\d+$')
    match = index_regex.search(string)
    if match is None:
        return None
    return int(match.group())


def input_int(
        s=None, ge=None, gt=None, le=None, lt=None, default=None, inset=None,
        raise_error=False, log=True):
    """Interactively prompt the user to enter an integer."""
    return _input_int_or_num(
        'int', s, ge, gt, le, lt, default, inset, raise_error, log)


def input_num(
        s=None, ge=None, gt=None, le=None, lt=None, default=None,
        raise_error=False, log=True):
    """Interactively prompt the user to enter a number."""
    return _input_int_or_num(
        'num', s, ge, gt, le, lt, default, None, raise_error,log)


def _input_int_or_num(
        type_str, s=None, ge=None, gt=None, le=None, lt=None, default=None,
        inset=None, raise_error=False, log=True):
    """Interactively prompt the user to enter an integer or number."""
    if type_str == 'int':
        if not test_ge_gt_le_lt(
                ge, gt, le, lt, is_int, '_input_int_or_num', raise_error, log):
            return None
    elif type_str == 'num':
        if not test_ge_gt_le_lt(
                ge, gt, le, lt, is_num, '_input_int_or_num', raise_error, log):
            return None
    else:
        illegal_value(
            type_str, 'type_str', '_input_int_or_num', raise_error, log)
        return None
    if default is not None:
        if not _is_int_or_num(
                default, type_str, raise_error=raise_error, log=log):
            return None
        if ge is not None and default < ge:
            illegal_combination(
                ge, 'ge', default, 'default', '_input_int_or_num', raise_error,
                log)
            return None
        if gt is not None and default <= gt:
            illegal_combination(
                gt, 'gt', default, 'default', '_input_int_or_num', raise_error,
                log)
            return None
        if le is not None and default > le:
            illegal_combination(
                le, 'le', default, 'default', '_input_int_or_num', raise_error,
                log)
            return None
        if lt is not None and default >= lt:
            illegal_combination(
                lt, 'lt', default, 'default', '_input_int_or_num', raise_error,
                log)
            return None
        default_string = f' [{default}]'
    else:
        default_string = ''
    if inset is not None:
        if (not isinstance(inset, (tuple, list))
                or any(not isinstance(i, int) for i in inset)):
            illegal_value(
                inset, 'inset', '_input_int_or_num', raise_error, log)
            return None
    v_range = f'{range_string_ge_gt_le_lt(ge, gt, le, lt)}'
    if v_range:
        v_range = f' {v_range}'
    if s is None:
        if type_str == 'int':
            print(f'Enter an integer{v_range}{default_string}: ')
        else:
            print(f'Enter a number{v_range}{default_string}: ')
    else:
        print(f'{s}{v_range}{default_string}: ')
    try:
        i = input()
        if isinstance(i, str) and not i:
            v = default
            print(f'{v}')
        else:
            v = literal_eval(i)
        if inset and v not in inset:
            raise ValueError(f'{v} not part of the set {inset}')
    except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError):
        v = None
    if not _is_int_or_num(v, type_str, ge, gt, le, lt):
        v = _input_int_or_num(
            type_str, s, ge, gt, le, lt, default, inset, raise_error, log)
    return v


def input_int_list(
        s=None, ge=None, le=None, split_on_dash=True, remove_duplicates=True,
        sort=True, raise_error=False, log=True):
    """
    Prompt the user to input a list of interger and split the entered
    string on any combination of commas, whitespaces, or dashes (when
    split_on_dash is True).
    e.g: '1 3,5-8 , 12 ' -> [1, 3, 5, 6, 7, 8, 12]

    remove_duplicates: removes duplicates if True (may also change the
        order)
    sort: sort in ascending order if True
    return None upon an illegal input
    """
    return _input_int_or_num_list(
        'int', s, ge, le, split_on_dash, remove_duplicates, sort, raise_error,
        log)


def input_num_list(
        s=None, ge=None, le=None, remove_duplicates=True, sort=True,
        raise_error=False, log=True):
    """
    Prompt the user to input a list of numbers and split the entered
    string on any combination of commas or whitespaces.
    e.g: '1.0, 3, 5.8, 12 ' -> [1.0, 3.0, 5.8, 12.0]

    remove_duplicates: removes duplicates if True (may also change the
        order)
    sort: sort in ascending order if True
    return None upon an illegal input
    """
    return _input_int_or_num_list(
        'num', s, ge, le, False, remove_duplicates, sort, raise_error, log)


def _input_int_or_num_list(
        type_str, s=None, ge=None, le=None, split_on_dash=True,
        remove_duplicates=True, sort=True, raise_error=False, log=True):
    # RV do we want a limit on max dimension?
    if type_str == 'int':
        if not test_ge_gt_le_lt(
                ge, None, le, None, is_int, 'input_int_or_num_list',
                raise_error, log):
            return None
    elif type_str == 'num':
        if not test_ge_gt_le_lt(
                ge, None, le, None, is_num, 'input_int_or_num_list',
                raise_error, log):
            return None
    else:
        illegal_value(type_str, 'type_str', '_input_int_or_num_list')
        return None
    v_range = f'{range_string_ge_gt_le_lt(ge=ge, le=le)}'
    if v_range:
        v_range = f' (each value in {v_range})'
    if s is None:
        print(f'Enter a series of integers{v_range}: ')
    else:
        print(f'{s}{v_range}: ')
    try:
        _list = string_to_list(input(), split_on_dash, remove_duplicates, sort)
    except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError):
        _list = None
    except:
        print('Unexpected error')
        raise
    if (not isinstance(_list, list) or any(
            not _is_int_or_num(v, type_str, ge=ge, le=le) for v in _list)):
        if split_on_dash:
            print('Invalid input: enter a valid set of dash/comma/whitespace '
                  'separated integers e.g. 1 3,5-8 , 12')
        else:
            print('Invalid input: enter a valid set of comma/whitespace '
                  'separated integers e.g. 1 3,5 8 , 12')
        _list = _input_int_or_num_list(
            type_str, s, ge, le, split_on_dash, remove_duplicates, sort,
            raise_error, log)
    return _list


def input_yesno(s=None, default=None):
    """Interactively prompt the user to enter a y/n question."""
    if default is not None:
        if not isinstance(default, str):
            illegal_value(default, 'default', 'input_yesno')
            return None
        if default.lower() in 'yes':
            default = 'y'
        elif default.lower() in 'no':
            default = 'n'
        else:
            illegal_value(default, 'default', 'input_yesno')
            return None
        default_string = f' [{default}]'
    else:
        default_string = ''
    if s is None:
        print(f'Enter yes or no{default_string}: ')
    else:
        print(f'{s}{default_string}: ')
    i = input()
    if isinstance(i, str) and not i:
        i = default
        print(f'{i}')
    if i is not None and i.lower() in 'yes':
        v = True
    elif i is not None and i.lower() in 'no':
        v = False
    else:
        print('Invalid input, enter yes or no')
        v = input_yesno(s, default)
    return v


def input_menu(items, default=None, header=None):
    """Interactively prompt the user to select from a menu."""
    if (not isinstance(items, (tuple, list))
            or any(not isinstance(i, str) for i in items)):
        illegal_value(items, 'items', 'input_menu')
        return None
    if default is not None:
        if not (isinstance(default, str) and default in items):
            logger.error(
                f'Invalid value for default ({default}), must be in {items}')
            return None
        default_string = f' [{1+items.index(default)}]'
    else:
        default_string = ''
    if header is None:
        print('Choose one of the following items '
              f'(1, {len(items)}){default_string}:')
    else:
        print(f'{header} (1, {len(items)}){default_string}:')
    for i, choice in enumerate(items):
        print(f'  {i+1}: {choice}')
    try:
        choice = input()
        if isinstance(choice, str) and not choice:
            choice = items.index(default)
            print(f'{1+choice}')
        else:
            choice = literal_eval(choice)
            if isinstance(choice, int) and 1 <= choice <= len(items):
                choice -= 1
            else:
                raise ValueError
    except (ValueError, TypeError, SyntaxError, MemoryError, RecursionError):
        choice = None
    except:
        print('Unexpected error')
        raise
    if choice is None:
        print(f'Invalid choice, enter a number between 1 and {len(items)}')
        choice = input_menu(items, default)
    return choice


def assert_no_duplicates_in_list_of_dicts(_list, raise_error=False):
    """
    Assert that there are no duplicates in a list of dictionaries.
    """
    if not isinstance(_list, list):
        illegal_value(
            _list, '_list', 'assert_no_duplicates_in_list_of_dicts',
            raise_error)
        return None
    if any(not isinstance(d, dict) for d in _list):
        illegal_value(
            _list, '_list', 'assert_no_duplicates_in_list_of_dicts',
            raise_error)
        return None
    if (len(_list) != len([dict(_tuple) for _tuple in
                          {tuple(sorted(d.items())) for d in _list}])):
        if raise_error:
            raise ValueError(f'Duplicate items found in {_list}')
        logger.error(f'Duplicate items found in {_list}')
        return None
    return _list


def assert_no_duplicate_key_in_list_of_dicts(_list, key, raise_error=False):
    """
    Assert that there are no duplicate keys in a list of dictionaries.
    """
    if not isinstance(key, str):
        illegal_value(
            key, 'key', 'assert_no_duplicate_key_in_list_of_dicts',
            raise_error)
        return None
    if not isinstance(_list, list):
        illegal_value(
            _list, '_list', 'assert_no_duplicate_key_in_list_of_dicts',
            raise_error)
        return None
    if any(isinstance(d, dict) for d in _list):
        illegal_value(
            _list, '_list', 'assert_no_duplicates_in_list_of_dicts',
            raise_error)
        return None
    keys = [d.get(key, None) for d in _list]
    if None in keys or len(set(keys)) != len(_list):
        if raise_error:
            raise ValueError(
                f'Duplicate or missing key ({key}) found in {_list}')
        logger.error(f'Duplicate or missing key ({key}) found in {_list}')
        return None
    return _list


def assert_no_duplicate_attr_in_list_of_objs(_list, attr, raise_error=False):
    """
    Assert that there are no duplicate attributes in a list of objects.
    """
    if not isinstance(attr, str):
        illegal_value(
            attr, 'attr', 'assert_no_duplicate_attr_in_list_of_objs',
            raise_error)
        return None
    if not isinstance(_list, list):
        illegal_value(
            _list, '_list', 'assert_no_duplicate_key_in_list_of_objs',
            raise_error)
        return None
    attrs = [getattr(obj, attr, None) for obj in _list]
    if None in attrs or len(set(attrs)) != len(_list):
        if raise_error:
            raise ValueError(
                f'Duplicate or missing attr ({attr}) found in {_list}')
        logger.error(f'Duplicate or missing attr ({attr}) found in {_list}')
        return None
    return _list


def file_exists_and_readable(f):
    """Check if a file exists and is readable."""
    if not os_path.isfile(f):
        raise ValueError(f'{f} is not a valid file')
    if not access(f, R_OK):
        raise ValueError(f'{f} is not accessible for reading')
    return f


def draw_mask_1d(
        ydata, xdata=None, label=None, ref_data=[],
        current_index_ranges=None, current_mask=None,
        select_mask=True, num_index_ranges_max=None,
        title=None, xlabel=None, ylabel=None,
        test_mode=False, return_figure=False):
    """Display a 2D plot and have the user select a mask.

    :param ydata: data array for which a mask will be constructed
    :type ydata: numpy.ndarray
    :param xdata: x-coordinates of the reference data, defaults to
        None
    :type xdata: numpy.ndarray, optional
    :param label: legend label for the reference data, defaults to
        None
    :type label: str, optional
    :param ref_data: a list of additional reference data to
        plot. Items in the list should be tuples of positional
        arguments and keyword arguments to unpack and pass directly to
        `matplotlib.axes.Axes.plot`, defaults to []
    :type ref_data: list[tuple[tuple, dict]]
    :param current_index_ranges: list of preselected index ranges to
        mask, defaults to None
    :type current_index_ranges: list[tuple[int, int]]
    :param current_mask: preselected boolean mask array, defaults to
        None
    :type current_mask: numpy.ndarray, optional
    :param select_mask: if True, user-selected ranges will be included
        when the returned mask is applied to `ydata`. If False, they
        will be excluded. Defaults to True.
    :type select_mask: bool, optional
    :param title: title for the displayed figure, defaults to None
    :type title: str, optional
    :param xlabel: label for the x-axis of the displayed figure,
        defaults to None
    :type xlabel: str, optional
    :param ylabel: label for the y-axis of the displayed figure,
        defaults to None
    :type ylabel: str, optional
    :param test_mode: if True, run as a non-interactive test
        case. Defaults to False
    :type test_mode: bool, optional
    :param return_figure: if True, also return a matplotlib figure of
        the drawn mask, defaults to False
    :type return_figure: bool, optional
    :return: a boolean mask array and the list of selected index
        ranges (and a matplotlib figure, if `return_figure` was True).
    :rtype: numpy.ndarray, list[tuple[int, int]] [, matplotlib.figure.Figure]
    """
    # RV make color blind friendly
    def draw_selections(
            ax, current_include, current_exclude, selected_index_ranges):
        """Draw the selections."""
        ax.clear()
        if title is not None:
            ax.set_title(title)
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        ax.plot(xdata, ydata, 'k', label=label)
        for data in ref_data:
            ax.plot(*data[0], **data[1])
        ax.legend()
        for low, upp in current_include:
            xlow = 0.5 * (xdata[max(0, low-1)]+xdata[low])
            xupp = 0.5 * (xdata[upp]+xdata[min(num_data-1, 1+upp)])
            ax.axvspan(xlow, xupp, facecolor='green', alpha=0.5)
        for low, upp in current_exclude:
            xlow = 0.5 * (xdata[max(0, low-1)]+xdata[low])
            xupp = 0.5 * (xdata[upp]+xdata[min(num_data-1, 1+upp)])
            ax.axvspan(xlow, xupp, facecolor='red', alpha=0.5)
        for low, upp in selected_index_ranges:
            xlow = 0.5 * (xdata[max(0, low-1)]+xdata[low])
            xupp = 0.5 * (xdata[upp]+xdata[min(num_data-1, 1+upp)])
            ax.axvspan(xlow, xupp, facecolor=selection_color, alpha=0.5)
        ax.get_figure().canvas.draw()

    def onclick(event):
        """Action taken on clicking the mouse button."""
        if event.inaxes in [fig.axes[0]]:
            selected_index_ranges.append(index_nearest_upp(xdata, event.xdata))

    def onrelease(event):
        """Action taken on releasing the mouse button."""
        if selected_index_ranges:
            if isinstance(selected_index_ranges[-1], int):
                if event.inaxes in [fig.axes[0]]:
                    event.xdata = index_nearest_low(xdata, event.xdata)
                    if selected_index_ranges[-1] <= event.xdata:
                        selected_index_ranges[-1] = \
                            (selected_index_ranges[-1], event.xdata)
                    else:
                        selected_index_ranges[-1] = \
                            (event.xdata, selected_index_ranges[-1])
                    draw_selections(
                        event.inaxes, current_include, current_exclude,
                        selected_index_ranges)
                else:
                    selected_index_ranges.pop(-1)

    def confirm_selection(event):
        """Action taken on hitting the confirm button."""
        plt.close()

    def clear_last_selection(event):
        """Action taken on hitting the clear button."""
        if selected_index_ranges:
            selected_index_ranges.pop(-1)
        else:
            while current_include:
                current_include.pop()
            while current_exclude:
                current_exclude.pop()
            selected_mask.fill(False)
        draw_selections(
            ax, current_include, current_exclude, selected_index_ranges)

    def update_mask(mask, selected_index_ranges, unselected_index_ranges):
        """Update the plot with the selected mask."""
        for low, upp in selected_index_ranges:
            selected_mask = np.logical_and(
                xdata >= xdata[low], xdata <= xdata[upp])
            mask = np.logical_or(mask, selected_mask)
        for low, upp in unselected_index_ranges:
            unselected_mask = np.logical_and(
                xdata >= xdata[low], xdata <= xdata[upp])
            mask[unselected_mask] = False
        return mask

    def update_index_ranges(mask):
        """
        Update the currently included index ranges (where mask = True).
        """
        current_include = []
        for i, m in enumerate(mask):
            if m:
                if (not current_include
                        or isinstance(current_include[-1], tuple)):
                    current_include.append(i)
            else:
                if current_include and isinstance(current_include[-1], int):
                    current_include[-1] = (current_include[-1], i-1)
        if current_include and isinstance(current_include[-1], int):
            current_include[-1] = (current_include[-1], num_data-1)
        return current_include

    # Check inputs
    ydata = np.asarray(ydata)
    if ydata.ndim > 1:
        logger.warning(f'Invalid ydata dimension ({ydata.ndim})')
        return None, None
    num_data = ydata.size
    if xdata is None:
        xdata = np.arange(num_data)
    else:
        xdata = np.asarray(xdata, dtype=np.float64)
        if xdata.ndim > 1 or xdata.size != num_data:
            logger.warning(f'Invalid xdata shape ({xdata.shape})')
            return None, None
        if not np.all(xdata[:-1] < xdata[1:]):
            logger.warning('Invalid xdata: must be monotonically increasing')
            return None, None
    if current_index_ranges is not None:
        if not isinstance(current_index_ranges, (tuple, list)):
            logger.warning(
                'Invalid current_index_ranges parameter '
                f'({current_index_ranges}, {type(current_index_ranges)})')
            return None, None
    if not isinstance(select_mask, bool):
        logger.warning(
            f'Invalid select_mask parameter ({select_mask}, '
            f'{type(select_mask)})')
        return None, None
    if num_index_ranges_max is not None:
        logger.warning(
            'num_index_ranges_max input not yet implemented in draw_mask_1d')
    if title is None:
        title = 'select ranges of data'
    elif not isinstance(title, str):
        illegal_value(title, 'title')
        title = ''

    if select_mask:
        selection_color = 'green'
    else:
        selection_color = 'red'

    # Set initial selected mask and the selected/unselected index
    #     ranges as needed
    selected_index_ranges = []
    unselected_index_ranges = []
    selected_mask = np.full(xdata.shape, False, dtype=bool)
    if current_index_ranges is None:
        if current_mask is None:
            if not select_mask:
                selected_index_ranges = [(0, num_data-1)]
                selected_mask = np.full(xdata.shape, True, dtype=bool)
        else:
            selected_mask = np.copy(np.asarray(current_mask, dtype=bool))
    if current_index_ranges is not None and current_index_ranges:
        current_index_ranges = sorted(list(current_index_ranges))
        for low, upp in current_index_ranges:
            if low > upp or low >= num_data or upp < 0:
                continue
            low = max(low, 0)
            upp = min(upp, num_data-1)
            selected_index_ranges.append((low, upp))
        selected_mask = update_mask(
            selected_mask, selected_index_ranges, unselected_index_ranges)
    if current_index_ranges is not None and current_mask is not None:
        selected_mask = np.logical_and(current_mask, selected_mask)
    if current_mask is not None:
        selected_index_ranges = update_index_ranges(selected_mask)

    # Set up range selections for display
    current_include = selected_index_ranges
    current_exclude = []
    selected_index_ranges = []
    if not current_include:
        if select_mask:
            current_exclude = [(0, num_data-1)]
        else:
            current_include = [(0, num_data-1)]
    else:
        if current_include[0][0] > 0:
            current_exclude.append((0, current_include[0][0]-1))
        for i in range(1, len(current_include)):
            current_exclude.append(
                (1+current_include[i-1][1], current_include[i][0]-1))
        if current_include[-1][1] < num_data-1:
            current_exclude.append((1+current_include[-1][1], num_data-1))

    # Set up matplotlib figure
    plt.close('all')
    fig, ax = plt.subplots()
    plt.subplots_adjust(bottom=0.2)
    draw_selections(
        ax, current_include, current_exclude, selected_index_ranges)

    if not test_mode:
        # Set up event handling for click-and-drag range selection
        cid_click = fig.canvas.mpl_connect('button_press_event', onclick)
        cid_release = fig.canvas.mpl_connect('button_release_event', onrelease)

        # Set up confirm / clear range selection buttons
        confirm_b = Button(plt.axes([0.75, 0.015, 0.15, 0.075]), 'Confirm')
        clear_b = Button(plt.axes([0.59, 0.015, 0.15, 0.075]), 'Clear')
        cid_confirm = confirm_b.on_clicked(confirm_selection)
        cid_clear = clear_b.on_clicked(clear_last_selection)

        # Show figure
        plt.show(block=True)

        # Disconnect callbacks when figure is closed
        fig.canvas.mpl_disconnect(cid_click)
        fig.canvas.mpl_disconnect(cid_release)
        confirm_b.disconnect(cid_confirm)
        clear_b.disconnect(cid_clear)

        # Remove buttons & readjust axes before returning a figure
        if return_figure:
            confirm_b.ax.remove()
            clear_b.ax.remove()
            plt.subplots_adjust(bottom=0.0)

    # Swap selection depending on select_mask
    if not select_mask:
        selected_index_ranges, unselected_index_ranges = \
            unselected_index_ranges, selected_index_ranges

    # Update the mask with the currently selected/unselected x-ranges
    selected_mask = update_mask(
        selected_mask, selected_index_ranges, unselected_index_ranges)

    # Update the currently included index ranges (where mask is True)
    current_include = update_index_ranges(selected_mask)

    if return_figure:
        return selected_mask, current_include, fig
    return selected_mask, current_include

def select_peaks(
        ydata, xdata, peak_locations,
        peak_labels=None,
        mask=None,
        pre_selected_peak_indices=[],
        return_sorted=True,
        title='', xlabel='', ylabel='',
        interactive=True):
    """
    Show a plot of the 1D data provided with user-selectable markers
    at the given locations. Return the locations of the markers that
    the user selected with their mouse interactions.

    :param ydata: 1D array of values to plot
    :type ydata: numpy.ndarray
    :param xdata: values of the independent dimension corresponding to
        ydata, defaults to None
    :type xdata: numpy.ndarray, optional
    :param peak_locations: locations of selectable markers in the same
        units as xdata.
    :type peak_locations: list
    :param peak_labels: list of annotations for each peak, defaults to
        None
    :type peak_labels: list[str], optional
    :param mask: boolean array representing a mask that will be
        applied to the data at some later point, defaults to None
    :type mask: np.ndarray, optional
    :param pre_selected_peak_indices: indices of markers that should
        already be selected when the figure shows up, defaults to []
    :type pre_selected_peak_indices: list[int], optional
    :param return_sorted: sort the indices of selected markers before
        returning (otherwise: return them in the same order that the
        user selected them), defaults to True
    :type return_sorted: bool, optional
    :param title: title for the plot, defaults to ''
    :type title: str, optional
    :param xlabel: x-axis label for the plot, defaults to ''
    :type xlabel: str, optional
    :param ylabel: y-axis label for the plot, defaults to ''
    :type ylabel: str, optional
    :param interactive: show the plot and allow user interactions with
        the matplotlib figure, defults to True
    :type interactive: bool, optional
    :return: the locations of the user-selected peaks
    :rtype: list
    """

    if ydata.size != xdata.size:
        raise ValueError('x and y data must have the same size')
    if mask is not None and mask.size != ydata.size:
        raise ValueError('mask must have the same size as data')


    excluded_peak_props = {
        'color': 'black', 'linestyle': '--','linewidth': 1,
        'marker': 10, 'markersize': 5, 'fillstyle': 'none'}
    included_peak_props = {
        'color': 'green', 'linestyle': '-', 'linewidth': 2,
        'marker': 10, 'markersize': 10, 'fillstyle': 'full'}
    masked_peak_props = {
        'color': 'gray', 'linestyle': ':', 'linewidth': 1}

    # Setup reference data & plot
    if mask is None:
        mask = np.full(ydata.shape, True, dtype=bool)
    fig, ax = plt.subplots(figsize=(11, 8.5))
    handles = ax.plot(xdata, ydata, label='Reference data')
    handles.append(mlines.Line2D(
        [], [], label='Excluded / unselected', **excluded_peak_props))
    handles.append(mlines.Line2D(
        [], [], label='Included / selected', **included_peak_props))
    handles.append(mlines.Line2D(
        [], [], label='In masked region (unselectable)', **masked_peak_props))
    ax.legend(handles=handles, loc='center right')
    ax.set(title=title, xlabel=xlabel, ylabel=ylabel)
    fig.tight_layout()

    # Plot a vertical line marker at each peak location
    peak_vlines = []
    x_indices = np.arange(ydata.size)
    if peak_labels is None:
        peak_labels = [''] * len(peak_locations)
    for i, (loc, lbl) in enumerate(zip(peak_locations, peak_labels)):
        nearest_index = np.searchsorted(xdata, loc)
        if nearest_index in x_indices[mask]:
            if i in pre_selected_peak_indices:
                peak_vline = ax.axvline(loc, **included_peak_props)
            else:
                peak_vline = ax.axvline(loc, **excluded_peak_props)
            peak_vline.set_picker(5)
        else:
            if i in pre_selected_peak_indices:
                logger.warning(
                    f'Pre-selected peak index {i} is in a masked region and '
                    'will not be selectable.')
                pre_selected_peak_indices.remove(i)
            peak_vline = ax.axvline(loc, **masked_peak_props)
        ax.text(loc, 1, lbl, ha='right', va='top', rotation=90,
                transform=ax.get_xaxis_transform())
        peak_vlines.append(peak_vline)

    # Indicate masked regions by gray-ing out the axes facecolor
    exclude_bounds = []
    for i, m in enumerate(mask):
        if not m:
            if (not exclude_bounds) or isinstance(exclude_bounds[-1], tuple):
                exclude_bounds.append(i)
        else:
            if exclude_bounds and isinstance(exclude_bounds[-1], int):
                exclude_bounds[-1] = (exclude_bounds[-1], i-1)
    if exclude_bounds and isinstance(exclude_bounds[-1], int):
        exclude_bounds[-1] = (exclude_bounds[-1], mask.size-1)
    for (low, upp) in exclude_bounds:
        xlow = xdata[low]
        xupp = xdata[upp]
        ax.axvspan(xlow, xupp, facecolor='gray', alpha=0.5)

    selected_peak_indices = pre_selected_peak_indices
    if interactive:
        # Setup interative peak selection
        def onpick(event):
            try:
                peak_index = peak_vlines.index(event.artist)
            except:
                pass
            else:
                peak_vline = event.artist
                if peak_index in selected_peak_indices:
                    peak_vline.set(**excluded_peak_props)
                    selected_peak_indices.remove(peak_index)
                else:
                    peak_vline.set(**included_peak_props)
                    selected_peak_indices.append(peak_index)
                plt.draw()
        cid_pick_peak = fig.canvas.mpl_connect('pick_event', onpick)

        # Setup "Confirm" button
        def confirm_selection(event):
            plt.close()
        plt.subplots_adjust(bottom=0.2)
        confirm_b = Button(plt.axes([0.75, 0.05, 0.15, 0.075]), 'Confirm')
        cid_confirm = confirm_b.on_clicked(confirm_selection)

        # Show figure for user interaction
        plt.show()

        # Disconnect all widget callbacks when figure is closed
        fig.canvas.mpl_disconnect(cid_pick_peak)
        confirm_b.disconnect(cid_confirm)

        # ...and remove the confirm button before returning the figure
        confirm_b.ax.remove()
        plt.subplots_adjust(bottom=0.0)

    selected_peaks = peak_locations[selected_peak_indices]
    if return_sorted:
        selected_peaks.sort()
    return selected_peaks, fig


def select_image_bounds(
        a, axis, low=None, upp=None, num_min=None, title='select array bounds',
        raise_error=False):
    """
    Interactively select the lower and upper data bounds for a 2D
    numpy array.
    """
    a = np.asarray(a)
    if a.ndim != 2:
        illegal_value(
            a.ndim, 'array dimension', location='select_image_bounds',
            raise_error=raise_error)
        return None
    if axis < 0 or axis >= a.ndim:
        illegal_value(
            axis, 'axis', location='select_image_bounds',
            raise_error=raise_error)
        return None
    low_save = low
    upp_save = upp
    num_min_save = num_min
    if num_min is None:
        num_min = 1
    else:
        if num_min < 2 or num_min > a.shape[axis]:
            logger.warning(
                'Invalid input for num_min in select_image_bounds, '
                'input ignored')
            num_min = 1
    if low is None:
        min_ = 0
        max_ = a.shape[axis]
        low_max = a.shape[axis]-num_min
        while True:
            if axis:
                quick_imshow(
                    a[:,min_:max_], title=title, aspect='auto',
                    extent=[min_,max_,a.shape[0],0])
            else:
                quick_imshow(
                    a[min_:max_,:], title=title, aspect='auto',
                    extent=[0,a.shape[1], max_,min_])
            zoom_flag = input_yesno(
                'Set lower data bound (y) or zoom in (n)?', 'y')
            if zoom_flag:
                low = input_int('    Set lower data bound', ge=0, le=low_max)
                break
            min_ = input_int('    Set lower zoom index', ge=0, le=low_max)
            max_ = input_int(
                '    Set upper zoom index', ge=min_+1, le=low_max+1)
    else:
        if not is_int(low, ge=0, le=a.shape[axis]-num_min):
            illegal_value(
                low, 'low', location='select_image_bounds',
                raise_error=raise_error)
            return None
    if upp is None:
        min_ = low+num_min
        max_ = a.shape[axis]
        upp_min = min_
        while True:
            if axis:
                quick_imshow(
                    a[:,min_:max_], title=title, aspect='auto',
                    extent=[min_,max_,a.shape[0],0])
            else:
                quick_imshow(
                    a[min_:max_,:], title=title, aspect='auto',
                    extent=[0,a.shape[1], max_,min_])
            zoom_flag = input_yesno(
                'Set upper data bound (y) or zoom in (n)?', 'y')
            if zoom_flag:
                upp = input_int(
                    '    Set upper data bound', ge=upp_min, le=a.shape[axis])
                break
            min_ = input_int(
                '    Set upper zoom index', ge=upp_min, le=a.shape[axis]-1)
            max_ = input_int(
                '    Set upper zoom index', ge=min_+1, le=a.shape[axis])
    else:
        if not is_int(upp, ge=low+num_min, le=a.shape[axis]):
            illegal_value(
                upp, 'upp', location='select_image_bounds',
                raise_error=raise_error)
            return None
    bounds = (low, upp)
    a_tmp = np.copy(a)
    a_tmp_max = a.max()
    if axis:
        a_tmp[:,bounds[0]] = a_tmp_max
        a_tmp[:,bounds[1]-1] = a_tmp_max
    else:
        a_tmp[bounds[0],:] = a_tmp_max
        a_tmp[bounds[1]-1,:] = a_tmp_max
    print(f'lower bound = {low} (inclusive)\nupper bound = {upp} (exclusive)')
    quick_imshow(a_tmp, title=title, aspect='auto')
    del a_tmp
    if not input_yesno('Accept these bounds (y/n)?', 'y'):
        bounds = select_image_bounds(
            a, axis, low=low_save, upp=upp_save, num_min=num_min_save,
            title=title)
    clear_imshow(title)
    return bounds


def select_one_image_bound(
        a, axis, bound=None, bound_name=None, title='select array bounds',
        default='y', raise_error=False):
    """
    Interactively select a data boundary for a 2D numpy array.
    """
    a = np.asarray(a)
    if a.ndim != 2:
        illegal_value(
            a.ndim, 'array dimension', location='select_one_image_bound',
            raise_error=raise_error)
        return None
    if axis < 0 or axis >= a.ndim:
        illegal_value(
            axis, 'axis', location='select_one_image_bound',
            raise_error=raise_error)
        return None
    if bound_name is None:
        bound_name = 'data bound'
    if bound is None:
        min_ = 0
        max_ = a.shape[axis]
        bound_max = a.shape[axis]-1
        while True:
            if axis:
                quick_imshow(
                    a[:,min_:max_], title=title, aspect='auto',
                    extent=[min_,max_,a.shape[0],0])
            else:
                quick_imshow(
                    a[min_:max_,:], title=title, aspect='auto',
                    extent=[0,a.shape[1], max_,min_])
            zoom_flag = input_yesno(
                f'Set {bound_name} (y) or zoom in (n)?', 'y')
            if zoom_flag:
                bound = input_int(f'    Set {bound_name}', ge=0, le=bound_max)
                clear_imshow(title)
                break
            min_ = input_int('    Set lower zoom index', ge=0, le=bound_max)
            max_ = input_int(
                '    Set upper zoom index', ge=min_+1, le=bound_max+1)

    elif not is_int(bound, ge=0, le=a.shape[axis]-1):
        illegal_value(
            bound, 'bound', location='select_one_image_bound',
            raise_error=raise_error)
        return None
    else:
        print(f'Current {bound_name} = {bound}')
    a_tmp = np.copy(a)
    a_tmp_max = a.max()
    if axis:
        a_tmp[:,bound] = a_tmp_max
    else:
        a_tmp[bound,:] = a_tmp_max
    quick_imshow(a_tmp, title=title, aspect='auto')
    del a_tmp
    if not input_yesno(f'Accept this {bound_name} (y/n)?', default):
        bound = select_one_image_bound(
            a, axis, bound_name=bound_name, title=title)
    clear_imshow(title)
    return bound


def clear_imshow(title=None):
    """Clear an image opened by quick_imshow()."""
    plt.ioff()
    if title is None:
        title = 'quick imshow'
    elif not isinstance(title, str):
        raise ValueError(f'Invalid parameter title ({title})')
    plt.close(fig=title)


def clear_plot(title=None):
    """Clear an image opened by quick_plot()."""
    plt.ioff()
    if title is None:
        title = 'quick plot'
    elif not isinstance(title, str):
        raise ValueError(f'Invalid parameter title ({title})')
    plt.close(fig=title)


def quick_imshow(
        a, title=None, path=None, name=None, save_fig=False, save_only=False,
        clear=True, extent=None, show_grid=False, grid_color='w',
        grid_linewidth=1, block=False, **kwargs):
    """Display a 2D image."""
    if title is not None and not isinstance(title, str):
        raise ValueError(f'Invalid parameter title ({title})')
    if path is not None and not isinstance(path, str):
        raise ValueError(f'Invalid parameter path ({path})')
    if not isinstance(save_fig, bool):
        raise ValueError(f'Invalid parameter save_fig ({save_fig})')
    if not isinstance(save_only, bool):
        raise ValueError(f'Invalid parameter save_only ({save_only})')
    if not isinstance(clear, bool):
        raise ValueError(f'Invalid parameter clear ({clear})')
    if not isinstance(block, bool):
        raise ValueError(f'Invalid parameter block ({block})')
    if not title:
        title = 'quick imshow'
    if name is None:
        ttitle = re_sub(r'\s+', '_', title)
        if path is None:
            path = f'{ttitle}.png'
        else:
            path = f'{path}/{ttitle}.png'
    else:
        if path is None:
            path = name
        else:
            path = f'{path}/{name}'
    if ('cmap' in kwargs and a.ndim == 3
            and (a.shape[2] == 3 or a.shape[2] == 4)):
        use_cmap = True
        if a.shape[2] == 4 and a[:,:,-1].min() != a[:,:,-1].max():
            use_cmap = False
        if any(
                a[i,j,0] != a[i,j,1] and a[i,j,0] != a[i,j,2]
                for i in range(a.shape[0])
                for j in range(a.shape[1])):
            use_cmap = False
        if use_cmap:
            a = a[:,:,0]
        else:
            logger.warning('Image incompatible with cmap option, ignore cmap')
            kwargs.pop('cmap')
    if extent is None:
        extent = (0, a.shape[1], a.shape[0], 0)
    if clear:
        try:
            plt.close(fig=title)
        except:
            pass
    if not save_only:
        if block:
            plt.ioff()
        else:
            plt.ion()
    plt.figure(title)
    plt.imshow(a, extent=extent, **kwargs)
    if show_grid:
        ax = plt.gca()
        ax.grid(color=grid_color, linewidth=grid_linewidth)
#    if title != 'quick imshow':
#        plt.title = title
    if save_only:
        plt.savefig(path)
        plt.close(fig=title)
    else:
        if save_fig:
            plt.savefig(path)
        if block:
            plt.show(block=block)


def quick_plot(
        *args, xerr=None, yerr=None, vlines=None, title=None, xlim=None,
        ylim=None, xlabel=None, ylabel=None, legend=None, path=None, name=None,
        show_grid=False, save_fig=False, save_only=False, clear=True,
        block=False, **kwargs):
    """Display a 2D line plot."""
    if title is not None and not isinstance(title, str):
        illegal_value(title, 'title', 'quick_plot')
        title = None
    if (xlim is not None and not isinstance(xlim, (tuple, list))
            and len(xlim) != 2):
        illegal_value(xlim, 'xlim', 'quick_plot')
        xlim = None
    if (ylim is not None and not isinstance(ylim, (tuple, list))
            and len(ylim) != 2):
        illegal_value(ylim, 'ylim', 'quick_plot')
        ylim = None
    if xlabel is not None and not isinstance(xlabel, str):
        illegal_value(xlabel, 'xlabel', 'quick_plot')
        xlabel = None
    if ylabel is not None and not isinstance(ylabel, str):
        illegal_value(ylabel, 'ylabel', 'quick_plot')
        ylabel = None
    if legend is not None and not isinstance(legend, (tuple, list)):
        illegal_value(legend, 'legend', 'quick_plot')
        legend = None
    if path is not None and not isinstance(path, str):
        illegal_value(path, 'path', 'quick_plot')
        return
    if not isinstance(show_grid, bool):
        illegal_value(show_grid, 'show_grid', 'quick_plot')
        return
    if not isinstance(save_fig, bool):
        illegal_value(save_fig, 'save_fig', 'quick_plot')
        return
    if not isinstance(save_only, bool):
        illegal_value(save_only, 'save_only', 'quick_plot')
        return
    if not isinstance(clear, bool):
        illegal_value(clear, 'clear', 'quick_plot')
        return
    if not isinstance(block, bool):
        illegal_value(block, 'block', 'quick_plot')
        return
    if title is None:
        title = 'quick plot'
    if name is None:
        ttitle = re_sub(r'\s+', '_', title)
        if path is None:
            path = f'{ttitle}.png'
        else:
            path = f'{path}/{ttitle}.png'
    else:
        if path is None:
            path = name
        else:
            path = f'{path}/{name}'
    if clear:
        try:
            plt.close(fig=title)
        except:
            pass
    args = unwrap_tuple(args)
    if depth_tuple(args) > 1 and (xerr is not None or yerr is not None):
        logger.warning('Error bars ignored form multiple curves')
    if not save_only:
        if block:
            plt.ioff()
        else:
            plt.ion()
    plt.figure(title)
    if depth_tuple(args) > 1:
        for y in args:
            plt.plot(*y, **kwargs)
    else:
        if xerr is None and yerr is None:
            plt.plot(*args, **kwargs)
        else:
            plt.errorbar(*args, xerr=xerr, yerr=yerr, **kwargs)
    if vlines is not None:
        if isinstance(vlines, (int, float)):
            vlines = [vlines]
        for v in vlines:
            plt.axvline(v, color='r', linestyle='--', **kwargs)
#    if vlines is not None:
#        for s in tuple(
#                ([x, x], list(plt.gca().get_ylim())) for x in vlines):
#            plt.plot(*s, color='red', **kwargs)
    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if show_grid:
        ax = plt.gca()
        ax.grid(color='k')  # , linewidth=1)
    if legend is not None:
        plt.legend(legend)
    if save_only:
        plt.savefig(path)
        plt.close(fig=title)
    else:
        if save_fig:
            plt.savefig(path)
        plt.show(block=block)



"""This subpackage contains a collection of
[pydantic](https://docs.pydantic.dev/latest/) models that are used by
`Processors` in [`CHAP.common`](CHAP.common.md) and elsewhere to
validate inputs.
"""



"""Utilities for interacting with scans using an SMB-style .par file
as input
"""

import csv
import json
import os

class ParFile():
    """Representation of a .par file

    :ivar par_file: name of the .par file
    :type par_file: str
    :ivar json_file: name of the .json file containing the key for
        column names of the .par file
    :type json_file: str
    :ivar spec_file: name of the SPEC data file associated with this
        .par file
    :type spec_file: str
    :ivar column_names: list of the names of each column in the par file
    :type column_names: list[str]
    :ivar data: a 2D array of the data in this .par file. 0th index:
        row. 1st index: column
    :type data: list[list]
    """
    def __init__(self, par_file, scann_col_name='SCAN_N'):
        self.par_file = str(par_file)
        self.json_file = self.par_file.replace('.par', '.json')
        self.spec_file = os.path.join(
            os.path.dirname(self.par_file), 'spec.log')

        with open(self.json_file) as json_file:
            columns = json.load(json_file)
        self.column_names = [None] * len(columns)
        for i, name in columns.items():
            self.column_names[int(i)] = name

        self.data = []
        with open(self.par_file) as par_file:
            reader = csv.reader(par_file, delimiter=' ')
            for row in reader:
                if len(row) == 0:
                    continue
                if row[0].startswith('#'):
                    continue
                row_data = []
                for value in row:
                    try:
                        value = int(value)
                    except ValueError:
                        try:
                            value = float(value)
                        except:
                            pass
                    row_data.append(value)
                self.data.append(row_data)

        self.scann_i = self.column_names.index(scann_col_name)
        self.scan_numbers = [data[self.scann_i] for data in self.data]

    def get_map(self, experiment_type, station, par_dims, other_dims=[]):
        """Return a map configuration based on this par file.

        :param experiment_type: name of the experiment type for the
            map that this .par file represents
        :type experiment_type: Literal['SAXSWAXS', 'EDD', 'XRF', 'TOMO']
        :param station: name of the station at which the data were
            collected
        :type station: Literal['id1a3','id3a','id3b']
        :param par_dims: list of dictionaries configuring the map's
            independent dimensions.
        :type par_dims: list[dict[str, str]]
        :param other_dims: a list of other dimensions to include in
            the returned MapConfig's independednt_dimensions. Use this
            if each scans in thhis par ile captured more than one
            frame of data. Defaults to []
        :type other_dims: list[dict[str,str]], optional
        :return: a map configuration
        :rtype: CHAP.common.models.map.MapConfig
        """
        import numpy as np
        from CHAP.common.models.map import MapConfig
        from CHAP.utils.scanparsers import SMBScanParser

        scanparser = SMBScanParser(self.spec_file, 1)
        good_scans = self.good_scan_numbers()
        map_config = {
            'title': scanparser.scan_name,
            'station': station, #scanparser.station,
            'experiment_type': experiment_type,
            'sample': {'name': scanparser.scan_name},
            'spec_scans': [
                {'spec_file': self.spec_file,
                 'scan_numbers': good_scans}],
            'independent_dimensions': [
                {'label': dim['label'],
                 'units': dim['units'],
                 'name': dim['name'],
                 'data_type': 'smb_par'}
                for dim in par_dims] + other_dims
        }
        map_config = MapConfig(**map_config)
        map_size = np.prod(map_config.shape)
        if map_size != len(good_scans):
            raise ValueError(
                f'{self.par_file} has {len(good_scans)} good scans, '
                + 'but size of map along '
                + ', '.join([dim['name'] for dim in par_dims + other_dims])
                + f' is {map_size}.')
        return map_config

    def good_scan_numbers(self, good_col_name='1/0'):
        """Return the numbers of scans marked with a "1" in the
        indicated "good" column of the .par file.
        
        :param good_col_name: the name of the "good" column of the par
            file, defaults to "1/0"
        :type good_col_name: str, optional
        :raises ValueError: if this .par file does not have a column
            with the same name as `good_col_name`
        :return: "good" scan numbers
        :rtype: list[int]
        """
        good_col_i = self.column_names.index(good_col_name)
        return [self.scan_numbers[i] for i in range(len(self.scan_numbers))
                if self.data[i][good_col_i] == 1]
        
    def get_values(self, column, scan_numbers=None):
        """Return values from a single column of the par file.

        :param column: the string name OR index of the column to return
            values for
        :type column: str or int
        :param scan_numbers: list of specific scan numbers to return
            values in the given column for (instead of the default
            behavior: return the entire column of values), defaults to
            None
        :type scan_numbers: list[int], optional
        :return: a list of values from a single column in the par file
        :rtype: list[object]
        """
        if isinstance(column, str):
            column_idx = self.column_names.index(column)
        elif isinstance(column, int):
            column_idx = column
        else:
            raise TypeError(f'column must be a str or int, not {type(column)}')

        column_data = [self.data[i][column_idx] for i in range(len(self.data))]
        if scan_numbers is not None:
            column_data = [column_data[self.scan_numbers.index(scan_n)] \
                           for scan_n in scan_numbers]
        return column_data

    def map_values(self, map_config, values):
        """Return a reshaped array of the 1D list `values` so that it
        matches up with the coordinates of `map_config`.

        :param map_config: the map configuration according to which
            values will be reshaped
        :type map_config: MapConfig
        :param values: a 1D list of values to reshape
        :type values: list or np.ndarray
        :return: reshaped array of values
        :rtype: np.ndarray
        """
        import numpy as np
        good_scans = self.good_scan_numbers()
        if len(values) != len(good_scans):
            raise ValueError('number of values provided ({len(values)}) does '
                             + 'not match the number of good scans in '
                             + f'{self.par_file} ({len(good_scans)})')
        n_map_points = np.prod(map_config.shape)
        if len(values) != n_map_points:
            raise ValueError(
                f'Cannot reshape {len(values)} values into an array of shape '
                + f'{map_config.shape}')

        map_values = np.empty(map_config.shape)
        for map_index in np.ndindex(map_config.shape):
            scans, scan_number, scan_step_index = \
                map_config.get_scan_step_index(map_index)
            value_index = good_scans.index(scan_number)
            map_values[map_index] = values[value_index]
        return map_values



"""The ChessAnalysisPipeline (CHAP) provides infrastructure to
construct and run X-ray data processing / analysis workflows using a
set of modular components. We call these components `PipelineItem`s
(subclassed into `Reader`s, `Processor`s, and `Writer`s). A `Pipeline`
uses a sequence of `PipelineItem`s to execute a data processing
workflow where the data returned by one `PipelineItem` becomes input
for the next one.

Many `PipelineItem`s can be shared by data processing workflows for
multiple different X-ray techniques, while others may be unique to
just a single technique. The `PipelineItem`s that are shared by many
techniques are organized in the `CHAP.common` subpackage.
`PipelineItem`s unique to a tomography workflow, for instance, are
organized in the `CHAP.tomo` subpackage.

[`CHAP.utils`](CHAP.utils.md) contains a
broad selection of utilities to assist in some common tasks that
appear in specific `Processor` implementations.
"""

from CHAP.reader import Reader
from CHAP.processor import Processor
from CHAP.writer import Writer



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : mnist_img.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: 
"""

import json
import gzip
import argparse
# from itertools import chain

import numpy as np
import matplotlib.pyplot as plt


def readImage(fname, fout, num_images=5, imgId=2):
    """
    Helper function to read MNIST image
    """
    image_size = 28
    with gzip.open(fname, 'r') as fstream:
        fstream.read(16)
        buf = fstream.read(image_size * image_size * num_images)
        data = np.frombuffer(buf, dtype=np.uint8).astype(np.float32)
        data = data.reshape(num_images, image_size, image_size, 1)
        image = np.asarray(data[imgId]).squeeze()
        plt.imsave(fout, image)
        print("read:", fname, "wrote:", fout, "image:", type(image), "shape:", image.shape)

def img2json(image):
    """
    Convert given image to JSON data format used by TFaaS
    """
    # values = [int(i) for i in list(chain.from_iterable(image))]
    # values = image.tolist()
    values = []
    for row in image.tolist():
        row = [int(i) for i in row]
        vals = [[i] for i in row]
        values.append(vals)
    # final values should be an array of elements, e.g. single image representation
    values = [values]
    keys = [str(i) for i in range(0, 10)]
    meta = {
        'keys': keys,
        'values': values,
        'model': 'mnist'
    }
    with open('img.json', 'w') as ostream:
        ostream.write(json.dumps(meta))


class OptionParser():
    def __init__(self):
        "User based option parser"
        fname = "train-images-idx3-ubyte.gz"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--fin", action="store",
            dest="fin", default=fname, help=f"Input MNIST file, default {fname}")
        self.parser.add_argument("--fout", action="store",
            dest="fout", default="img.png", help="Output image fila name, default img.png")
        self.parser.add_argument("--nimages", action="store",
            dest="nimages", default=5, help="number of images to read, default 5")
        self.parser.add_argument("--imgid", action="store",
            dest="imgid", default=2, help="image index to use from nimages, default 2 (number 4)")

def main():
    """
    main function to produce image file from mnist dataset.
    MNIST dataset can be downloaded from
    curl -O http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz
    """
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    num_images = int(opts.nimages)
    imgId = int(opts.imgid)
    img = readImage(opts.fin, opts.fout, num_images, imgId)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : tfaas_client.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: simple python client to communicate with TFaaS server
"""

# system modules
import os
import sys
import pwd
import ssl
import json
import binascii
import argparse
import itertools
import mimetypes
if  sys.version_info < (2, 7):
    raise Exception("TFaaS client requires python 2.7 or greater")
# python 3
if  sys.version.startswith('3.'):
    import urllib.request as urllib2
    import urllib.parse as urllib
    import http.client as httplib
    import http.cookiejar as cookielib
else:
    import mimetools
    import urllib
    import urllib2
    import httplib
    import cookielib

TFAAS_CLIENT = 'tfaas-client/1.1::python/%s.%s' % sys.version_info[:2]

class OptionParser():
    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--url", action="store",
            dest="url", default="", help="TFaaS URL")
        self.parser.add_argument("--upload", action="store",
            dest="upload", default="", help="upload model to TFaaS")
        self.parser.add_argument("--bundle", action="store",
            dest="bundle", default="", help="upload bundle ML files to TFaaS")
        self.parser.add_argument("--predict", action="store",
            dest="predict", default="", help="fetch prediction from TFaaS")
        self.parser.add_argument("--image", action="store",
            dest="image", default="", help="fetch prediction for given image")
        self.parser.add_argument("--model", action="store",
            dest="model", default="", help="TF model to use")
        self.parser.add_argument("--delete", action="store",
            dest="delete", default="", help="delete model in TFaaS")
        self.parser.add_argument("--models", action="store_true",
            dest="models", default=False, help="show existing models in TFaaS")
        self.parser.add_argument("--verbose", action="store_true",
            dest="verbose", default=False, help="verbose output")
        msg  = 'specify private key file name, default $X509_USER_PROXY'
        self.parser.add_argument("--key", action="store",
                               default=x509(), dest="ckey", help=msg)
        msg  = 'specify private certificate file name, default $X509_USER_PROXY'
        self.parser.add_argument("--cert", action="store",
                               default=x509(), dest="cert", help=msg)
        default_ca = os.environ.get("X509_CERT_DIR")
        if not default_ca or not os.path.exists(default_ca):
            default_ca = "/etc/grid-security/certificates"
            if not os.path.exists(default_ca):
                default_ca = ""
        if default_ca:
            msg = 'specify CA path, default currently is %s' % default_ca
        else:
            msg = 'specify CA path; defaults to system CAs.'
        self.parser.add_argument("--capath", action="store",
                               default=default_ca, dest="capath", help=msg)
        msg  = 'specify number of retries upon busy DAS server message'

class HTTPSClientAuthHandler(urllib2.HTTPSHandler):
    """
    Simple HTTPS client authentication class based on provided
    key/ca information
    """
    def __init__(self, key=None, cert=None, capath=None, level=0):
        if  level > 0:
            urllib2.HTTPSHandler.__init__(self, debuglevel=1)
        else:
            urllib2.HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert
        self.capath = capath

    def https_open(self, req):
        """Open request method"""
        #Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.get_connection, req)

    def get_connection(self, host, timeout=300):
        """Connection method"""
        if  self.key and self.cert and not self.capath:
            return httplib.HTTPSConnection(host, key_file=self.key,
                                                cert_file=self.cert)
        elif self.cert and self.capath:
            context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
            context.load_verify_locations(capath=self.capath)
            context.load_cert_chain(self.cert)
            return httplib.HTTPSConnection(host, context=context)
        return httplib.HTTPSConnection(host)

def x509():
    "Helper function to get x509 either from env or tmp file"
    proxy = os.environ.get('X509_USER_PROXY', '')
    if  not proxy:
        proxy = '/tmp/x509up_u%s' % pwd.getpwuid( os.getuid() ).pw_uid
        if  not os.path.isfile(proxy):
            return ''
    return proxy

def check_auth(key):
    "Check if user runs das_client with key/cert and warn users to switch"
    if  not key:
        msg  = "WARNING: tfaas_client is running without user credentials/X509 proxy, create proxy via 'voms-proxy-init -voms cms -rfc'"
        print(msg)

def fullpath(path):
    "Expand path to full path"
    if  path and path[0] == '~':
        path = path.replace('~', '')
        path = path[1:] if path[0] == '/' else path
        path = os.path.join(os.environ['HOME'], path)
    return path

def choose_boundary():
    """
    Helper function to replace deprecated mimetools.choose_boundary
    https://stackoverflow.com/questions/27099290/where-is-mimetools-choose-boundary-function-in-python3
    https://docs.python.org/2.7/library/mimetools.html?highlight=choose_boundary#mimetools.choose_boundary
    >>> mimetools.choose_boundary()
    '192.168.1.191.502.42035.1678979116.376.1'
    """
    # we will return any random string
    import uuid
    return str(uuid.uuid4())

# credit: https://pymotw.com/2/urllib2/#uploading-files
class MultiPartForm(object):
    """Accumulate the data to be used when posting a form."""

    def __init__(self):
        self.form_fields = []
        self.files = []
        if  sys.version.startswith('3.'):
            self.boundary = choose_boundary()
        else:
            self.boundary = mimetools.choose_boundary()
        return
    
    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary

    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
        return

    def add_file(self, fieldname, filename, fileHandle, mimetype=None):
        """Add a file to be uploaded."""
        body = fileHandle.read()
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        if mimetype == 'application/octet-stream':
            body = binascii.b2a_base64(body)
#         if isinstance(body, bytes):
#             body = body.decode("utf-8") 
        self.files.append((fieldname, filename, mimetype, body))
        return
    
    def __str__(self):
        """Return a string representing the form data, including attached files."""
        # Build a list of lists, each containing "lines" of the
        # request.  Each part is separated by a boundary string.
        # Once the list is built, return a string where each
        # line is separated by '\r\n'.  
        parts = []
        part_boundary = '--' + self.boundary
        
        # Add the form fields
        parts.extend(
            [ part_boundary,
              'Content-Disposition: form-data; name="%s"' % name,
              '',
              value,
            ]
            for name, value in self.form_fields
            )
        
        # Add the files to upload
        # here we use form-data content disposition instead of file one
        # since this is how we define handlers in our Go server
        # for more info see: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition
        parts.extend(
            [ part_boundary,
              'Content-Disposition: form-data; name="%s"; filename="%s"' % \
                 (field_name, filename),
              'Content-Type: %s' % content_type,
              '',
              body,
            ]
            for field_name, filename, content_type, body in self.files
            )
        
        # Flatten the list and add closing boundary marker,
        # then return CR+LF separated data
        flattened = list(itertools.chain(*parts))
        flattened.append('--' + self.boundary + '--')
        flattened.append('')
        return '\r\n'.join(flattened)

def models(host, verbose=None, ckey=None, cert=None, capath=None):
    "models API shows models from TFaaS server"
    url = host + '/models'
    client = '%s (%s)' % (TFAAS_CLIENT, os.environ.get('USER', ''))
    headers = {"Accept": "application/json", "User-Agent": client}
    if verbose:
        print("URL   : %s" % url)
    encoded_data = json.dumps({})
    return getdata(url, headers, encoded_data, ckey, cert, capath, verbose, 'GET')

def delete(host, model, verbose=None, ckey=None, cert=None, capath=None):
    "delete API deletes given model in TFaaS server"
    url = host + '/delete'
    client = '%s (%s)' % (TFAAS_CLIENT, os.environ.get('USER', ''))
    headers = {"User-Agent": client}
    if verbose:
        print("URL   : %s" % url)
        print("model : %s" % model)
    form = MultiPartForm()
    form.add_field('model', model)
    edata = str(form)
    headers['Content-length'] = len(edata)
    headers['Content-Type'] = form.get_content_type()
    return getdata(url, headers, edata, ckey, cert, capath, verbose, method='DELETE')

def bundle(host, ifile, verbose=None, ckey=None, cert=None, capath=None):
    "bundle API uploads given bundle model files to TFaaS server"
    url = host + '/upload'
    client = '%s (%s)' % (TFAAS_CLIENT, os.environ.get('USER', ''))
    headers = {"User-Agent": client, "Content-Encoding": "gzip", "Content-Type": "application/octet-stream"}
    data = open(ifile, 'rb').read()
    return getdata(url, headers, data, ckey, cert, capath, verbose)

def upload(host, ifile, verbose=None, ckey=None, cert=None, capath=None):
    "upload API uploads given model to TFaaS server"
    url = host + '/upload'
    client = '%s (%s)' % (TFAAS_CLIENT, os.environ.get('USER', ''))
    headers = {"User-Agent": client}
    params = json.load(open(ifile))
    if verbose:
        print("URL   : %s" % url)
        print("ifile : %s" % ifile)
        print("params: %s" % json.dumps(params))

    form = MultiPartForm()
    for key in params.keys():
        if key in ['model', 'labels', 'params']:
            flag = 'r'
            if key == 'model':
                flag = 'rb'
            name = params[key]
            form.add_file(key, name, fileHandle=open(name, flag))
        else:
            form.add_field(key, params[key])
    edata = str(form)
    headers['Content-length'] = len(edata)
    headers['Content-Type'] = form.get_content_type()
    headers['Content-Encoding'] = 'base64'
    return getdata(url, headers, edata, ckey, cert, capath, verbose)

def predict(host, ifile, model, verbose=None, ckey=None, cert=None, capath=None):
    "predict API get predictions from TFaaS server"
    url = host + '/json'
    client = '%s (%s)' % (TFAAS_CLIENT, os.environ.get('USER', ''))
    headers = {"Accept": "application/json", "User-Agent": client}
    params = json.load(open(ifile))
    if model: # overwrite model name in given input file
        params['model'] = model
    if verbose:
        print("URL   : %s" % url)
        print("ifile : %s" % ifile)
        print("params: %s" % json.dumps(params))
    encoded_data = json.dumps(params)
    return getdata(url, headers, encoded_data, ckey, cert, capath, verbose)

def predictImage(host, ifile, model, verbose=None, ckey=None, cert=None, capath=None):
    "predict API get predictions from TFaaS server"
    url = host + '/image'
    client = '%s (%s)' % (TFAAS_CLIENT, os.environ.get('USER', ''))
    headers = {"Accept": "application/json", "User-Agent": client}
    if verbose:
        print("URL   : %s" % url)
        print("ifile : %s" % ifile)
        print("model : %s" % model)
    form = MultiPartForm()
#     form.add_file('image', ifile, fileHandle=open(ifile, 'r'))
    form.add_file('image', ifile, fileHandle=open(ifile, 'rb'))
    form.add_field('model', model)
    edata = str(form)
    headers['Content-length'] = len(edata)
    headers['Content-Type'] = form.get_content_type()
    return getdata(url, headers, edata, ckey, cert, capath, verbose)

def getdata(url, headers, encoded_data, ckey, cert, capath, verbose=None, method='POST'):
    "helper function to use in predict/upload APIs, it place given URL call to the server"
    debug = 1 if verbose else 0
    req = urllib2.Request(url=url, headers=headers, data=encoded_data)
    if method == 'DELETE':
        req.get_method = lambda: 'DELETE'
    elif method == 'GET':
        req = urllib2.Request(url=url, headers=headers)
    if  ckey and cert:
        ckey = fullpath(ckey)
        cert = fullpath(cert)
        http_hdlr  = HTTPSClientAuthHandler(ckey, cert, capath, debug)
    elif cert and capath:
        cert = fullpath(cert)
        http_hdlr  = HTTPSClientAuthHandler(ckey, cert, capath, debug)
    else:
        http_hdlr  = urllib2.HTTPHandler(debuglevel=debug)
    proxy_handler  = urllib2.ProxyHandler({})
    cookie_jar     = cookielib.CookieJar()
    cookie_handler = urllib2.HTTPCookieProcessor(cookie_jar)
    data = {}
    try:
        opener = urllib2.build_opener(http_hdlr, proxy_handler, cookie_handler)
        fdesc = opener.open(req)
        if url.endswith('json'):
            data = json.load(fdesc)
        else:
            data = fdesc.read()
        fdesc.close()
    except urllib2.HTTPError as error:
        print(error.read())
        sys.exit(1)
    if url.endswith('json'):
        return json.dumps(data)
    return data

def main():
    "Main function"
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    check_auth(opts.ckey)
    res = ''
    if opts.upload:
        res = upload(opts.url, opts.upload, opts.verbose, opts.ckey, opts.cert, opts.capath)
    if opts.bundle:
        res = bundle(opts.url, opts.bundle, opts.verbose, opts.ckey, opts.cert, opts.capath)
    elif opts.delete:
        res = delete(opts.url, opts.delete, opts.verbose, opts.ckey, opts.cert, opts.capath)
    elif opts.models:
        res = models(opts.url, opts.verbose, opts.ckey, opts.cert, opts.capath)
    elif opts.predict:
        res = predict(opts.url, opts.predict, opts.model, opts.verbose, opts.ckey, opts.cert, opts.capath)
    elif opts.image:
        res = predictImage(opts.url, opts.image, opts.model, opts.verbose, opts.ckey, opts.cert, opts.capath)
    if res:
        print(res)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
File       : reader.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: generic Reader module

Define a generic `Reader` object.
"""

# system modules
import argparse
from inspect import getfullargspec
import logging
from sys import modules
from time import time

# local modules
from CHAP.pipeline import PipelineItem


class Reader(PipelineItem):
    """Generic file reader.

    The job of any `Reader` in a `Pipeline` is to provide data stored
    in a file to the next `PipelineItem`. Note that a `Reader` used on
    its own disrupts the flow of data in a `Pipeline` -- it does not
    receive or pass along any data returned by the previous
    `PipelineItem`.
    """

    def read(self, filename):
        """Read and return the contents of `filename` as text

        :param filename: Name of file to read from
        :type filename: str
        :return: entire contents of the file
        :rtype: str
        """

        if not filename:
            self.logger.warning(
                'No file name is given, will skip read operation')
            return None

        with open(filename) as file:
            data = file.read()
        return data


class OptionParser():
    """User based option parser"""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--filename', action='store',
            dest='filename', default='', help='Input file')
        self.parser.add_argument(
            '--reader', action='store',
            dest='reader', default='Reader', help='Reader class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function"""

    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.reader
    try:
        reader_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported reader {cls_name}')
        raise

    reader = reader_cls()
    reader.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    reader.logger.addHandler(log_handler)
    data = reader.read(filename=opts.filename)

    print(f'Reader {reader} reads from {opts.filename}, data {data}')


if __name__ == '__main__':
    main()



"""This subpackage contains `PipelineItems` unique to EDD data
processing workflows.
"""
# from CHAP.edd.reader import
from CHAP.edd.processor import (DiffractionVolumeLengthProcessor,
                                MCACeriaCalibrationProcessor,
                                MCADataProcessor,
                                StrainAnalysisProcessor)
# from CHAP.edd.writer import

from CHAP.common import MapProcessor



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : ktrain.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Keras based ML network to train over MNIST dataset
"""

# system modules
import os
import sys
import json
import gzip
import pickle
import argparse

# third-party modules
import numpy as np
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras import backend as K
from tensorflow.python.tools import saved_model_utils


def modelGraph(model_dir):
    """
    Provide input/output names used by TF Graph along with graph itself
    The code is based on TF saved_model_cli.py script.
    """
    input_names = []
    output_names = []
    tag_sets = saved_model_utils.get_saved_model_tag_sets(model_dir)
    for tag_set in sorted(tag_sets):
        print('%r' % ', '.join(sorted(tag_set)))
        meta_graph_def = saved_model_utils.get_meta_graph_def(model_dir, tag_set[0])
        for key in meta_graph_def.signature_def.keys():
            meta = meta_graph_def.signature_def[key]
            if hasattr(meta, 'inputs') and hasattr(meta, 'outputs'):
                inputs = meta.inputs
                outputs = meta.outputs
                input_signatures = list(meta.inputs.values())
                input_names = [signature.name for signature in input_signatures]
                if len(input_names) > 0:
                    output_signatures = list(meta.outputs.values())
                    output_names = [signature.name for signature in output_signatures]
    return input_names, output_names, meta_graph_def

def readData(fin, num_classes):
    """
    Helper function to read MNIST data and provide it to
    upstream code, e.g. to the training layer
    """
    # Load the data and split it between train and test sets
#     (x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()
    f = gzip.open(fin, 'rb')
    if sys.version_info < (3,):
        mnist_data = pickle.load(f)
    else:
        mnist_data = pickle.load(f, encoding='bytes')
    f.close()
    (x_train, y_train), (x_test, y_test) = mnist_data

    # Scale images to the [0, 1] range
    x_train = x_train.astype("float32") / 255
    x_test = x_test.astype("float32") / 255
    # Make sure images have shape (28, 28, 1)
    x_train = np.expand_dims(x_train, -1)
    x_test = np.expand_dims(x_test, -1)
    print("x_train shape:", x_train.shape)
    print(x_train.shape[0], "train samples")
    print(x_test.shape[0], "test samples")


    # convert class vectors to binary class matrices
    y_train = keras.utils.to_categorical(y_train, num_classes)
    y_test = keras.utils.to_categorical(y_test, num_classes)
    return x_train, y_train, x_test, y_test


def train(fin, fout=None, model_name=None, epochs=1, batch_size=128, h5=False):
    """
    train function for MNIST
    """
    # Model / data parameters
    num_classes = 10
    input_shape = (28, 28, 1)

    # create ML model
    model = keras.Sequential(
        [
            keras.Input(shape=input_shape),
            layers.Conv2D(32, kernel_size=(3, 3), activation="relu"),
            layers.MaxPooling2D(pool_size=(2, 2)),
            layers.Conv2D(64, kernel_size=(3, 3), activation="relu"),
            layers.MaxPooling2D(pool_size=(2, 2)),
            layers.Flatten(),
            layers.Dropout(0.5),
            layers.Dense(num_classes, activation="softmax"),
        ]
    )

    model.summary()
    print("model input", model.input, type(model.input), model.input.__dict__)
    print("model output", model.output, type(model.output), model.output.__dict__)
    model.compile(loss="categorical_crossentropy", optimizer="adam", metrics=["accuracy"])

    # train model
    x_train, y_train, x_test, y_test = readData(fin, num_classes)
    model.fit(x_train, y_train, batch_size=batch_size, epochs=epochs, validation_split=0.1)

    # evaluate trained model
    score = model.evaluate(x_test, y_test, verbose=0)
    print("Test loss:", score[0])
    print("Test accuracy:", score[1])
    print("save model to", fout)
    writer(fout, model_name, model, input_shape, h5)

def writer(fout, model_name, model, input_shape, h5=False):
    """
    Writer provide write function for given model
    """
    if not fout:
        return
    model.save(fout)
    if h5:
        model.save('{}/{}'.format(fout, h5), save_format='h5')
    pbModel = '{}/saved_model.pb'.format(fout)
    pbtxtModel = '{}/saved_model.pbtxt'.format(fout)
    convert(pbModel, pbtxtModel)

    # get meta-data information about our ML model
    input_names, output_names, model_graph = modelGraph(model_name)
    print("### input", input_names)
    print("### output", output_names)
    # ML uses (28,28,1) shape, i.e. 28x28 black-white images
    # if we'll use color images we'll use shape (28, 28, 3)
    img_channels = input_shape[2]  # last item represent number of colors
    meta = {'name': model_name,
            'model': 'saved_model.pb',
            'labels': 'labels.txt',
            'img_channels': img_channels,
            'input_name': input_names[0].split(':')[0],
            'output_name': output_names[0].split(':')[0],
            'input_node': model.input.name,
            'output_node': model.output.name
    }
    with open(fout+'/params.json', 'w') as ostream:
        ostream.write(json.dumps(meta))
    with open(fout+'/labels.txt', 'w') as ostream:
        for i in range(0, 10):
            ostream.write(str(i)+'\n')
    with open(fout + '/model.graph', 'wb') as ostream:
        ostream.write(model_graph.SerializeToString())

def convert(fin, fout):
    """
    convert input model.pb into output model.pbtxt
    Based on internet search:
    - https://www.tensorflow.org/guide/saved_model
    - https://www.programcreek.com/python/example/123317/tensorflow.core.protobuf.saved_model_pb2.SavedModel
    """
    import google.protobuf
    from tensorflow.core.protobuf import saved_model_pb2
    import tensorflow as tf

    saved_model = saved_model_pb2.SavedModel()

    with open(fin, 'rb') as f:
        saved_model.ParseFromString(f.read())
        
    with open(fout, 'w') as f:
        f.write(google.protobuf.text_format.MessageToString(saved_model))


class OptionParser():
    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--fin", action="store",
            dest="fin", default="", help="Input MNIST file")
        self.parser.add_argument("--fout", action="store",
            dest="fout", default="", help="Output models area")
        self.parser.add_argument("--model", action="store",
            dest="model", default="mnist", help="model name")
        self.parser.add_argument("--epochs", action="store",
            dest="epochs", default=1, help="number of epochs to use in ML training")
        self.parser.add_argument("--batch_size", action="store",
            dest="batch_size", default=128, help="batch size to use in training")
        self.parser.add_argument("--h5", action="store",
            dest="h5", default="mnist", help="h5 model file name")

def main():
    "Main function"
    optmgr  = OptionParser()
    opts = optmgr.parser.parse_args()
    train(opts.fin, opts.fout,
          model_name=opts.model,
          epochs=opts.epochs,
          batch_size=opts.batch_size,
          h5=opts.h5)

if __name__ == '__main__':
    main()



"""
File       : runner.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description:
"""

# system modules
import argparse
import logging
import os
from yaml import safe_load

# local modules
from CHAP.pipeline import Pipeline


class RunConfig():
    """Representation of Pipeline run configuration."""
    opts = {'root': os.getcwd(),
            'profile': False,
            'interactive': False,
            'log_level': 'INFO',
            'inputdir': '.',
            'outputdir': '.'}

    def __init__(self, config={}):
        """RunConfig constructor

        :param config: Pipeline configuration options
        :type config: dict
        """
        # System modules
        from tempfile import NamedTemporaryFile

        for opt in self.opts:
            setattr(self, opt, config.get(opt, self.opts[opt]))

        # Check if root exists (create it if not) and is readable
        if not os.path.isdir(self.root):
            os.mkdir(self.root)
        if not os.access(self.root, os.R_OK):
            raise OSError('root directory is not accessible for reading '
                          f'({self.root})')

        # Check if input exists and is readable
        if not os.path.isabs(self.inputdir):
            self.inputdir = os.path.realpath(
                os.path.join(self.root, self.inputdir))
        if not os.path.isdir(self.inputdir):
            raise OSError(f'input directory does not exist ({self.inputdir})')
        if not os.access(self.inputdir, os.R_OK):
            raise OSError('input directory is not accessible for reading '
                          f'({self.inputdir})')

        # Check if output exists (create it if not) and is writable
        if not os.path.isabs(self.outputdir):
            self.outputdir = os.path.realpath(
                os.path.join(self.root, self.outputdir))
        if not os.path.isdir(self.outputdir):
            os.mkdir(self.outputdir)
        try:
            tmpfile = NamedTemporaryFile(dir=self.outputdir)
        except:
            raise OSError('output directory is not accessible for writing '
                          f'({self.outputdir})')

        self.log_level = self.log_level.upper()

def parser():
    """Return an argument parser for the `CHAP` CLI. This parser has
    one argument: the input CHAP configuration file.
    """
    parser = argparse.ArgumentParser(prog='PROG')
    parser.add_argument(
        'config', action='store', default='',
        help='Input configuration file')
    return parser

def main():
    """Main function"""
    args = parser().parse_args()

    # read input config file
    configfile = args.config
    with open(configfile) as file:
        config = safe_load(file)
    run_config = RunConfig(config.get('config', {}))
    pipeline_config = config.get('pipeline', [])

    # profiling setup
    if run_config.profile:
        from cProfile import runctx  # python profiler
        from pstats import Stats     # profiler statistics
        cmd = 'runner(run_config, pipeline_config)'
        runctx(cmd, globals(), locals(), 'profile.dat')
        info = Stats('profile.dat')
        info.sort_stats('cumulative')
        info.print_stats()
    else:
        runner(run_config, pipeline_config)

def runner(run_config, pipeline_config):
    """Main runner funtion

    :param run_config: CHAP run configuration
    :type run_config: RunConfig
    :param pipeline_config: CHAP Pipeline configuration
    :type pipeline_config: dict
    """
    # logging setup
    logger, log_handler = setLogger(run_config.log_level)
    logger.info(f'Input pipeline configuration: {pipeline_config}\n')

    # run pipeline
    run(pipeline_config,
        run_config.inputdir, run_config.outputdir, run_config.interactive,
        logger, run_config.log_level, log_handler)

def setLogger(log_level="INFO"):
    """
    Helper function to set CHAP logger

    :param log_level: logger level, default INFO
    """
    logger = logging.getLogger(__name__)
    log_level = getattr(logging, log_level.upper())
    logger.setLevel(log_level)
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    logger.addHandler(log_handler)
    return logger, log_handler

def run(pipeline_config,
        inputdir=None, outputdir=None, interactive=False,
        logger=None, log_level=None, log_handler=None):
    """
    Run given pipeline_config

    :param pipeline_config: CHAP pipeline config
    """
    objects = []
    kwds = []
    for item in pipeline_config:
        # load individual object with given name from its module
        kwargs = {'interactive': interactive,
                  'inputdir': inputdir,
                  'outputdir': outputdir}
        if isinstance(item, dict):
            name = list(item.keys())[0]
            # Combine the "interactive" command line argument with the object's keywords
            # giving precedence of "interactive" in the latter
            kwargs = {**kwargs, **item[name]}
        else:
            name = item
        if "users" in name:
            # load users module. This is required in CHAPaaS which can
            # have common area for users module. Otherwise, we will be
            # required to have invidual user's PYTHONPATHs to load user
            # processors.
            try:
                import users
            except ImportError:
                if logger:
                    logger.error(f'Unable to load {name}')
                continue
            clsName = name.split('.')[-1]
            modName = '.'.join(name.split('.')[:-1])
            module = __import__(modName, fromlist=[clsName])
            obj = getattr(module, clsName)()
        else:
            modName, clsName = name.split('.')
            module = __import__(f'CHAP.{modName}', fromlist=[clsName])
            obj = getattr(module, clsName)()
        if log_level:
            obj.logger.setLevel(log_level)
        if log_handler:
            obj.logger.addHandler(log_handler)
        if logger:
            logger.info(f'Loaded {obj}')
        objects.append(obj)
        kwds.append(kwargs)
    pipeline = Pipeline(objects, kwds)
    if log_level:
        pipeline.logger.setLevel(log_level)
    if log_handler:
        pipeline.logger.addHandler(log_handler)
    if logger:
        logger.info(f'Loaded {pipeline} with {len(objects)} items\n')
        logger.info(f'Calling "execute" on {pipeline}')
    pipeline.execute()


if __name__ == '__main__':
    main()



# System modules
from functools import cache, lru_cache
import os
from typing import Literal, Optional, Union

# Third party modules
import numpy as np
from pydantic import (BaseModel,
                      conint,
                      conlist,
                      constr,
                      FilePath,
                      PrivateAttr,
                      validator)
from pyspec.file.spec import FileSpec


class Sample(BaseModel):
    """Class representing a sample metadata configuration.

    :ivar name: The name of the sample.
    :type name: str
    :ivar description: A description of the sample.
    :type description: Optional[str]
    """
    name: constr(min_length=1)
    description: Optional[str]


class SpecScans(BaseModel):
    """Class representing a set of scans from a single SPEC file.

    :ivar spec_file: Path to the SPEC file.
    :type spec_file: str
    :ivar scan_numbers: List of scan numbers to use.
    :type scan_numbers: list[int]
    """
    spec_file: FilePath
    scan_numbers: conlist(item_type=conint(gt=0), min_items=1)

    @validator('spec_file', allow_reuse=True)
    def validate_spec_file(cls, spec_file):
        """Validate the specified SPEC file.

        :param spec_file: Path to the SPEC file.
        :type spec_file: str
        :raises ValueError: If the SPEC file is invalid.
        :return: Absolute path to the SPEC file, if it is valid.
        :rtype: str
        """
        try:
            spec_file = os.path.abspath(spec_file)
            sspec_file = FileSpec(spec_file)
        except:
            raise ValueError(f'Invalid SPEC file {spec_file}')
        return spec_file

    @validator('scan_numbers', allow_reuse=True)
    def validate_scan_numbers(cls, scan_numbers, values):
        """Validate the specified list of scan numbers.

        :param scan_numbers: List of scan numbers.
        :type scan_numbers: list of int
        :param values: Dictionary of values for all fields of the model.
        :type values: dict
        :raises ValueError: If a specified scan number is not found in
            the SPEC file.
        :return: List of scan numbers.
        :rtype: list of int
        """
        spec_file = values.get('spec_file')
        if spec_file is not None:
            spec_scans = FileSpec(spec_file)
            for scan_number in scan_numbers:
                scan = spec_scans.get_scan_by_number(scan_number)
                if scan is None:
                    raise ValueError(
                        f'No scan number {scan_number} in {spec_file}')
        return scan_numbers

    @property
    def scanparsers(self):
        """A list of `ScanParser`s for each of the scans specified by
        the SPEC file and scan numbers belonging to this instance of
        `SpecScans`
        """
        return [self.get_scanparser(scan_no) for scan_no in self.scan_numbers]

    def get_scanparser(self, scan_number):
        """This method returns a `ScanParser` for the specified scan
        number in the specified SPEC file.

        :param scan_number: Scan number to get a `ScanParser` for
        :type scan_number: int
        :return: `ScanParser` for the specified scan number
        :rtype: ScanParser
        """
        return get_scanparser(self.spec_file, scan_number)

    def get_index(self, scan_number:int, scan_step_index:int, map_config):
        """This method returns a tuple representing the index of a
        specific step in a specific SPEC scan within a map.

        :param scan_number: Scan number to get index for
        :type scan_number: int
        :param scan_step_index: Scan step index to get index for
        :type scan_step_index: int
        :param map_config: Map configuration to get index for
        :type map_config: MapConfig
        :return: Index for the specified scan number and scan step
            index within the specified map configuration
        :rtype: tuple
        """
        index = ()
        for independent_dimension in map_config.independent_dimensions:
            coordinate_index = list(
                map_config.coords[independent_dimension.label]).index(
                    independent_dimension.get_value(
                        self, scan_number, scan_step_index))
            index = (coordinate_index, *index)
        return index

    def get_detector_data(self,
                          detectors:list,
                          scan_number:int,
                          scan_step_index:int):
        """Return the raw data from the specified detectors at the
        specified scan number and scan step index.

        :param detectors: List of detector prefixes to get raw data for
        :type detectors: list[str]
        :param scan_number: Scan number to get data for
        :type scan_number: int
        :param scan_step_index: Scan step index to get data for
        :type scan_step_index: int
        :return: Data from the specified detectors for the specified
            scan number and scan step index
        :rtype: list[np.ndarray]
        """
        return get_detector_data(
            tuple([detector.prefix for detector in detectors]),
            self.spec_file,
            scan_number,
            scan_step_index)


@cache
def get_available_scan_numbers(spec_file:str):
    scans = FileSpec(spec_file).scans
    scan_numbers = list(scans.keys())
    return scan_numbers


@cache
def get_scanparser(spec_file:str, scan_number:int):
    if scan_number not in get_available_scan_numbers(spec_file):
        return None
    return ScanParser(spec_file, scan_number)


@lru_cache(maxsize=10)
def get_detector_data(
        detector_prefixes:tuple,
        spec_file:str,
        scan_number:int,
        scan_step_index:int):
    detector_data = []
    scanparser = get_scanparser(spec_file, scan_number)
    for prefix in detector_prefixes:
        image_data = scanparser.get_detector_data(prefix, scan_step_index)
        detector_data.append(image_data)
    return detector_data


class PointByPointScanData(BaseModel):
    """Class representing a source of raw scalar-valued data for which
    a value was recorded at every point in a `MapConfig`.

    :ivar label: A user-defined label for referring to this data in
        the NeXus file and in other tools.
    :type label: str
    :ivar units: The units in which the data were recorded.
    :type units: str
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['spec_motor', 'scan_column', 'smb_par']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    """
    label: constr(min_length=1)
    units: constr(strip_whitespace=True, min_length=1)
    data_type: Literal['spec_motor', 'scan_column', 'smb_par']
    name: constr(strip_whitespace=True, min_length=1)

    @validator('label')
    def validate_label(cls, label):
        """Validate that the supplied `label` does not conflict with
        any of the values for `label` reserved for certain data needed
        to perform corrections.

        :param label: The value of `label` to validate
        :type label: str
        :raises ValueError: If `label` is one of the reserved values.
        :return: The original supplied value `label`, if it is
            allowed.
        :rtype: str
        """
        if ((not issubclass(cls,CorrectionsData))
                and label in CorrectionsData.reserved_labels()):
            raise ValueError(
                f'{cls.__name__}.label may not be any of the following '
                f'reserved values: {CorrectionsData.reserved_labels()}')
        return label

    def validate_for_station(self, station:str):
        """Validate this instance of `PointByPointScanData` for a
        certain choice of station (beamline).

        :param station: The name of the station (in 'idxx' format).
        :type station: str
        :raises TypeError: If the station is not compatible with the
            value of the `data_type` attribute for this instance of
            PointByPointScanData.
        :return: None
        :rtype: None
        """
        if (station.lower() not in ('id1a3', 'id3a')
                and self.data_type == 'smb_par'):
            raise TypeError(
                f'{self.__class__.__name__}.data_type may not be "smb_par" '
                f'when station is "{station}"')

    def validate_for_spec_scans(
            self, spec_scans:list[SpecScans],
            scan_step_index:Union[Literal['all'],int] = 'all'):
        """Validate this instance of `PointByPointScanData` for a list
        of `SpecScans`.

        :param spec_scans: A list of `SpecScans` whose raw data will
            be checked for the presence of the data represented by
            this instance of `PointByPointScanData`
        :type spec_scans: list[SpecScans]
        :param scan_step_index: A specific scan step index to validate,
            defaults to `'all'`.
        :type scan_step_index: Union[Literal['all'],int], optional
        :raises RuntimeError: If the data represented by this instance of
            `PointByPointScanData` is missing for the specified scan steps.
        :return: None
        :rtype: None
        """
        for scans in spec_scans:
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                if scan_step_index == 'all':
                    scan_step_index_range = range(scanparser.spec_scan_npts)
                else:
                    scan_step_index_range = range(
                        scan_step_index, 1+scan_step_index)
                for index in scan_step_index_range:
                    try:
                        self.get_value(scans, scan_number, index)
                    except:
                        raise RuntimeError(
                            f'Could not find data for {self.name} '
                            f'(data_type "{self.data_type}") '
                            f'on scan number {scan_number} '
                            f'for index {index} '
                            f'in spec file {scans.spec_file}')

    def get_value(self, spec_scans:SpecScans,
                  scan_number:int, scan_step_index:int=0):
        """Return the value recorded for this instance of
        `PointByPointScanData` at a specific scan step.

        :param spec_scans: An instance of `SpecScans` in which the
            requested scan step occurs.
        :type spec_scans: SpecScans
        :param scan_number: The number of the scan in which the
            requested scan step occurs.
        :type scan_number: int
        :param scan_step_index: The index of the requested scan step.
        :type scan_step_index: int
        :return: The value recorded of the data represented by this
            instance of `PointByPointScanData` at the scan step
            requested
        :rtype: float
        """
        if self.data_type == 'spec_motor':
            return get_spec_motor_value(spec_scans.spec_file,
                                        scan_number,
                                        scan_step_index,
                                        self.name)
        if self.data_type == 'scan_column':
            return get_spec_counter_value(spec_scans.spec_file,
                                          scan_number,
                                          scan_step_index,
                                          self.name)
        if self.data_type == 'smb_par':
            return get_smb_par_value(spec_scans.spec_file,
                                     scan_number,
                                     self.name)
        return None


@cache
def get_spec_motor_value(spec_file:str, scan_number:int,
                         scan_step_index:int, spec_mnemonic:str):
    """Return the value recorded for a SPEC motor at a specific scan
    step.

    :param spec_file: Location of a SPEC file in which the requested
        scan step occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param scan_step_index: The index of the requested scan step.
    :type scan_step_index: int
    :param spec_mnemonic: The menmonic of a SPEC motor.
    :type spec_mnemonic: str
    :return: The value of the motor at the scan step requested
    :rtype: float
    """
    scanparser = get_scanparser(spec_file, scan_number)
    if (hasattr(scanparser, 'spec_scan_motor_mnes')
            and spec_mnemonic in scanparser.spec_scan_motor_mnes):
        motor_i = scanparser.spec_scan_motor_mnes.index(spec_mnemonic)
        if scan_step_index >= 0:
            scan_step = np.unravel_index(
                scan_step_index,
                scanparser.spec_scan_shape,
                order='F')
            motor_value = \
                scanparser.spec_scan_motor_vals[motor_i][scan_step[motor_i]]
        else:
            motor_value = scanparser.spec_scan_motor_vals[motor_i]
    else:
        motor_value = scanparser.get_spec_positioner_value(spec_mnemonic)
    return motor_value


@cache
def get_spec_counter_value(spec_file:str, scan_number:int,
                           scan_step_index:int, spec_column_label:str):
    """Return the value recorded for a SPEC counter at a specific scan
    step.

    :param spec_file: Location of a SPEC file in which the requested
        scan step occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param scan_step_index: The index of the requested scan step.
    :type scan_step_index: int
    :param spec_column_label: The label of a SPEC data column.
    :type spec_column_label: str
    :return: The value of the counter at the scan step requested
    :rtype: float
    """
    scanparser = get_scanparser(spec_file, scan_number)
    if scan_step_index >= 0:
        return scanparser.spec_scan_data[spec_column_label][scan_step_index]
    return scanparser.spec_scan_data[spec_column_label]


@cache
def get_smb_par_value(spec_file:str, scan_number:int, par_name:str):
    """Return the value recorded for a specific scan in SMB-tyle .par
    file.

    :param spec_file: Location of a SPEC file in which the requested
        scan step occurs.
    :type spec_scans: str
    :param scan_number: The number of the scan in which the requested
        scan step occurs.
    :type scan_number: int
    :param par_name: The name of the column in the .par file
    :type par_name: str
    :return: The value of the .par file value for  the scan requested.
    :rtype: float
    """
    scanparser = get_scanparser(spec_file, scan_number)
    return scanparser.pars[par_name]


def validate_data_source_for_map_config(data_source, values):
    """Confirm that an instance of PointByPointScanData is valid for
    the station and scans provided by a map configuration dictionary.

    :param data_source: the input object to validate
    :type data_source: PintByPointScanData
    :param values: the map configuration dictionary
    :type values: dict
    :raises Exception: if `data_source` cannot be validated for
        `values`.
    :return: `data_source`, iff it is valid.
    :rtype: PointByPointScanData
    """
    if data_source is not None:
        import_scanparser(values.get('station'), values.get('experiment_type'))
        data_source.validate_for_station(values.get('station'))
        data_source.validate_for_spec_scans(values.get('spec_scans'))
    return data_source


class IndependentDimension(PointByPointScanData):
    """Class representing the source of data to identify the
    coordinate values along one dimension of a `MapConfig`

    :ivar label: A user-defined label for referring to this data in
        the NeXus file and in other tools.
    :type label: str
    :ivar units: The units in which the data were recorded.
    :type units: str
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['spec_motor', 'scan_column', 'smb_par']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    :param start: Sarting index for slicing all datasets of a
        `MapConfig` along this axis, defaults to 0
    :type start: int, optional
    :param end: Ending index for slicing all datasets of a `MapConfig`
        along this axis, defaults to the total number of unique values
        along this axis in the associated `MapConfig`
    :type end: int, optional
    :param step: Step for slicing all datasets of a `MapConfig` along
        this axis, defaults to 1
    :type step: int, optional
    """
    start: Optional[conint(ge=0)] = 0
    end: Optional[int] = None
    step: Optional[conint(gt=0)] = 1

#    @validator('step')
#    def validate_step(cls, value):
#        if value == 0 :
#            raise ValueError('slice step cannot be zero')
#        return value


class CorrectionsData(PointByPointScanData):
    """Class representing the special instances of
    `PointByPointScanData` that are used by certain kinds of
    `CorrectionConfig` tools.

    :ivar label: One of the reserved values required by
        `CorrectionConfig`, `'presample_intensity'`,
        `'postsample_intensity'`, or `'dwell_time_actual'`.
    :type label: Literal['presample_intensity',
                         'postsample_intensity',
                         'dwell_time_actual']
    :ivar units: The units in which the data were recorded.
    :type units: str
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['scan_column', 'smb_par']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    """
    label: Literal['presample_intensity',
                   'postsample_intensity',
                   'dwell_time_actual']
    data_type: Literal['scan_column','smb_par']

    @classmethod
    def reserved_labels(cls):
        """Return a list of all the labels reserved for
        corrections-related scalar data.

        :return: A list of reserved labels
        :rtype: list[str]
        """
        return list(cls.__fields__['label'].type_.__args__)


class PresampleIntensity(CorrectionsData):
    """Class representing a source of raw data for the intensity of
    the beam that is incident on the sample.

    :ivar label: Must be `"presample_intensity"`
    :type label: Literal["presample_intensity"]
    :ivar units: Must be `"counts"`
    :type units: Literal["counts"]
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['scan_column', 'smb_par']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    """
    label: Literal['presample_intensity'] = 'presample_intensity'
    units: Literal['counts'] = 'counts'


class PostsampleIntensity(CorrectionsData):
    """Class representing a source of raw data for the intensity of
    the beam that has passed through the sample.

    :ivar label: Must be `"postsample_intensity"`
    :type label: Literal["postsample_intensity"]
    :ivar units: Must be `"counts"`
    :type units: Literal["counts"]
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['scan_column', 'smb_par']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    """
    label: Literal['postsample_intensity'] = 'postsample_intensity'
    units: Literal['counts'] = 'counts'


class DwellTimeActual(CorrectionsData):
    """Class representing a source of raw data for the actual dwell
    time at each scan point in SPEC (with some scan types, this value
    can vary slightly point-to-point from the dwell time specified in
    the command).

    :ivar label: Must be `"dwell_time_actual"`
    :type label: Literal["dwell_time_actual"]
    :ivar units: Must be `"counts"`
    :type units: Literal["counts"]
    :ivar data_type: Represents how these data were recorded at time
        of data collection.
    :type data_type: Literal['scan_column', 'smb_par']
    :ivar name: Represents the name with which these raw data were
        recorded at time of data collection.
    :type name: str
    """
    label: Literal['dwell_time_actual'] = 'dwell_time_actual'
    units: Literal['s'] = 's'


class SpecConfig(BaseModel):
    """Class representing the raw data for one or more SPEC scans.

    :ivar station: The name of the station at which the data was
        collected.
    :type station: Literal['id1a3','id3a','id3b']
    :ivar spec_scans: A list of the SPEC scans that compose the set.
    :type spec_scans: list[SpecScans]
    """
    station: Literal['id1a3','id3a','id3b']
    experiment_type: Literal['SAXSWAXS', 'EDD', 'XRF', 'TOMO']
    spec_scans: conlist(item_type=SpecScans, min_items=1)

    @validator('experiment_type')
    def validate_experiment_type(cls, value, values):
        """Ensure values for the station and experiment_type fields are
        compatible
        """
        station = values.get('station')
        if station == 'id1a3':
            allowed_experiment_types = ['SAXSWAXS', 'EDD', 'TOMO']
        elif station == 'id3a':
            allowed_experiment_types = ['EDD', 'TOMO']
        elif station == 'id3b':
            allowed_experiment_types = ['SAXSWAXS', 'XRF', 'TOMO']
        else:
            allowed_experiment_types = []
        if value not in allowed_experiment_types:
            raise ValueError(
                f'For station {station}, allowed experiment types are '
                f'{", ".join(allowed_experiment_types)}. '
                f'Supplied experiment type {value} is not allowed.')
        import_scanparser(station, value)
        return value


class MapConfig(BaseModel):
    """Class representing an experiment consisting of one or more SPEC
    scans.

    :ivar title: The title for the map configuration.
    :type title: str
    :ivar station: The name of the station at which the map was
        collected.
    :type station: Literal['id1a3','id3a','id3b']
    :ivar spec_scans: A list of the SPEC scans that compose the map.
    :type spec_scans: list[SpecScans]
    :ivar independent_dimensions: A list of the sources of data
        representing the raw values of each independent dimension of
        the map.
    :type independent_dimensions: list[PointByPointScanData]
    :ivar presample_intensity: A source of point-by-point presample
        beam intensity data. Required when applying a CorrectionConfig
        tool.
    :type presample_intensity: Optional[PresampleIntensity]
    :ivar dwell_time_actual: A source of point-by-point actual dwell
        times for SPEC scans. Required when applying a
        CorrectionConfig tool.
    :type dwell_time_actual: Optional[DwellTimeActual]
    :ivar presample_intensity: A source of point-by-point postsample
        beam intensity data. Required when applying a CorrectionConfig
        tool with `correction_type="flux_absorption"` or
        `correction_type="flux_absorption_background"`.
    :type presample_intensity: Optional[PresampleIntensity]
    :ivar scalar_data: A list of the sources of data representing
        other scalar raw data values collected at each point on the
        map. In the NeXus file representation of the map, datasets for
        these values will be included.
    :type scalar_values: Optional[list[PointByPointScanData]]
    """
    title: constr(strip_whitespace=True, min_length=1)
    station: Literal['id1a3','id3a','id3b']
    experiment_type: Literal['SAXSWAXS', 'EDD', 'XRF', 'TOMO']
    sample: Sample
    spec_scans: conlist(item_type=SpecScans, min_items=1)
    independent_dimensions: conlist(item_type=IndependentDimension,
                                    min_items=1)
    presample_intensity: Optional[PresampleIntensity]
    dwell_time_actual: Optional[DwellTimeActual]
    postsample_intensity: Optional[PostsampleIntensity]
    scalar_data: Optional[list[PointByPointScanData]] = []
    _coords: dict = PrivateAttr()

    _validate_independent_dimensions = validator(
        'independent_dimensions',
        each_item=True,
        allow_reuse=True)(validate_data_source_for_map_config)
    _validate_presample_intensity = validator(
        'presample_intensity',
        allow_reuse=True)(validate_data_source_for_map_config)
    _validate_dwell_time_actual = validator(
        'dwell_time_actual',
        allow_reuse=True)(validate_data_source_for_map_config)
    _validate_postsample_intensity = validator(
        'postsample_intensity',
        allow_reuse=True)(validate_data_source_for_map_config)
    _validate_scalar_data = validator(
        'scalar_data',
        each_item=True,
        allow_reuse=True)(validate_data_source_for_map_config)

    @validator('experiment_type')
    def validate_experiment_type(cls, value, values):
        """Ensure values for the station and experiment_type fields are
        compatible
        """
        station = values.get('station')
        if station == 'id1a3':
            allowed_experiment_types = ['SAXSWAXS', 'EDD', 'TOMO']
        elif station == 'id3a':
            allowed_experiment_types = ['EDD', 'TOMO']
        elif station == 'id3b':
            allowed_experiment_types = ['SAXSWAXS', 'XRF', 'TOMO']
        else:
            allowed_experiment_types = []
        if value not in allowed_experiment_types:
            raise ValueError(
                f'For station {station}, allowed experiment types are '
                f'{", ".join(allowed_experiment_types)}. '
                f'Supplied experiment type {value} is not allowed.')
        return value

    @property
    def coords(self):
        """Return a dictionary of the values of each independent
        dimension across the map.

        :returns: A dictionary ofthe map's  coordinate values.
        :rtype: dict[str,list[float]]
        """
        try:
            coords = self._coords
        except:
            coords = {}
            for dim in self.independent_dimensions:
                coords[dim.label] = []
                for scans in self.spec_scans:
                    for scan_number in scans.scan_numbers:
                        scanparser = scans.get_scanparser(scan_number)
                        for scan_step_index in range(
                                scanparser.spec_scan_npts):
                            coords[dim.label].append(dim.get_value(
                                    scans, scan_number, scan_step_index))
                coords[dim.label] = np.unique(coords[dim.label])
                if dim.end is None:
                    dim.end = len(coords[dim.label])
                coords[dim.label] = coords[dim.label][slice(
                    dim.start, dim.end, dim.step)]
            self._coords = coords
        return coords

    @property
    def dims(self):
        """Return a tuple of the independent dimension labels for the
        map.
        """
        return [point_by_point_scan_data.label
                for point_by_point_scan_data
                in self.independent_dimensions[::-1]]

    @property
    def shape(self):
        """Return the shape of the map -- a tuple representing the
        number of unique values of each dimension across the map.
        """
        return tuple([len(values) for key,values in self.coords.items()][::-1])

    @property
    def all_scalar_data(self):
        """Return a list of all instances of `PointByPointScanData`
        for which this map configuration will collect dataset-like
        data (as opposed to axes-like data).

        This will be any and all of the items in the
        corrections-data-related fields, as well as any additional
        items in the optional `scalar_data` field.
        """
        return [getattr(self, label, None)
                for label in CorrectionsData.reserved_labels()
                if getattr(self, label, None) is not None] + self.scalar_data

    def get_scan_step_index(self, map_index):
        """Return parameters to identify a single SPEC scan step that
        corresponds to the map point at the index provided.

        :param map_index: The index of a map point to identify as a
            specific SPEC scan step index
        :type map_index: tuple
        :return: A `SpecScans` configuration, scan number, and scan
            step index
        :rtype: tuple[SpecScans, int, int]
        """
        coords = {dim: self.coords[dim][i]
            for dim,i in zip( self.dims, map_index)}
        for scans in self.spec_scans:
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                for scan_step_index in range(scanparser.spec_scan_npts):
                    _coords = {dim.label:dim.get_value(
                                   scans, scan_number, scan_step_index)
                               for dim in self.independent_dimensions}
                    if _coords == coords:
                        return scans, scan_number, scan_step_index
        raise RuntimeError(f'Unable to match coordinates {coords}')

    def get_value(self, data, map_index):
        """Return the raw data collected by a single device at a
        single point in the map.

        :param data: The device configuration to return a value of raw
            data for
        :type data: PointByPointScanData
        :param map_index: The map index to return raw data for
        :type map_index: tuple
        :return: Raw data value
        """
        scans, scan_number, scan_step_index = \
            self.get_scan_step_index(map_index)
        return data.get_value(scans, scan_number, scan_step_index)

    def get_detector_data(self, detector_name, map_index):
        """Return detector data collected by this map.

        :param detector_name: Name of the detector for which to return
            data. Usually the value of the detector's EPICS
            areaDetector prefix macro, $P.
        :type detector_name: str
        :param map_index: The map index to return detector data for.
        :return: One frame of raw detector data
        :rtype: np.ndarray
        """
        scans, scan_number, scan_step_index = \
            self.get_scan_step_index(map_index)
        scanparser = scans.get_scanparser(scan_number)
        return scanparser.get_detector_data(detector_name, scan_step_index)


def import_scanparser(station, experiment):
    """Given the name of a CHESS station and experiment type, import
    the corresponding subclass of `ScanParser` as `ScanParser`.

    :param station: The station name ("IDxx", not the beamline acronym)
    :type station: str
    :param experiment: The experiment type
    :type experiment: Literal["SAXSWAXS","EDD","XRF","Tomo","Powder"]
    :return: None
    """

    station = station.lower()
    experiment = experiment.lower()

    # Local modules
    if station in ('id1a3', 'id3a'):
        if experiment in ('saxswaxs', 'powder'):
            from CHAP.utils.scanparsers \
                import SMBLinearScanParser as ScanParser
        elif experiment == 'edd':
            from CHAP.utils.scanparsers \
                import SMBMCAScanParser as ScanParser
        elif experiment == 'tomo':
            from CHAP.utils.scanparsers \
                import SMBRotationScanParser as ScanParser
        else:
            raise ValueError(
                f'Invalid experiment type for station {station}: {experiment}')
    elif station == 'id3b':
        if experiment == 'saxswaxs':
            from CHAP.utils.scanparsers \
                import FMBSAXSWAXSScanParser as ScanParser
        elif experiment == 'tomo':
            from CHAP.utils.scanparsers \
                import FMBRotationScanParser as ScanParser
        elif experiment == 'xrf':
            from CHAP.utils.scanparsers \
                import FMBXRFScanParser as ScanParser
        else:
            raise ValueError(
                f'Invalid experiment type for station {station}: {experiment}')
    else:
        raise ValueError(f'Invalid station: {station}')

    globals()['ScanParser'] = ScanParser



#!/usr/bin/env python
"""
File       : reader.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Module for Writers used in multiple experiment-specific
             workflows.
"""

# system modules
from os.path import (
    isfile,
    splitext,
)
from sys import modules
from time import time

# local modules
from CHAP import Reader


class BinaryFileReader(Reader):
    """Reader for binary files"""
    def read(self, filename):
        """Return a content of a given file name

        :param filename: name of the binart file to read from
        :return: the content of `filename`
        :rtype: binary
        """

        with open(filename, 'rb') as file:
            data = file.read()
        return data


class SpecReader(Reader):
    """Reader for CHESS SPEC scans"""
    def read(self, filename=None, spec_config=None, detector_names=[]):
        """Take a SPEC configuration filename or dictionary and return
        the raw data as an NXentry.

        :param filename: name of file with the SPEC configuration to
            read and pass onto the constructor of
            `CHAP.common.models.map.SpecConfig`
        :type filename: str, optional
        :param spec_config: SPEC configuration to be passed directly
            to the constructor of
            `CHAP.common.models.map.SpecConfig`
        :type spec_config: dict, optional
        :param detector_names: Detector prefixes to include raw data
            for in the returned NXentry
        :type detector_names: list[str]
        :return: Data from the SPEC configuration provided
        :rtype: nexusformat.nexus.NXentry
        """
        from json import dumps
        from nexusformat.nexus import (NXcollection,
                                       NXdata,
                                       NXentry,
                                       NXfield)
        import numpy as np
        from CHAP.common.models.map import SpecConfig

        if filename is not None:
            if spec_config is not None:
                raise RuntimeError('Specify either filename or spec_config '
                                   'in common.SpecReader, not both')
            # Read the map configuration from file
            if not isfile(filename):
                raise OSError(f'input file does not exist ({filename})')
            extension = splitext(filename)[1]
            if extension in ('.yml', '.yaml'):
                reader = YAMLReader()
            else:
                raise RuntimeError('input file has a non-implemented '
                                   f'extension ({filename})')
            spec_config = reader.read(filename)
        elif not isinstance(spec_config, dict):
            raise RuntimeError('Invalid parameter spec_config in '
                               f'common.SpecReader ({spec_config})')

        # Validate the SPEC configuration provided by constructing a
        # SpecConfig
        spec_config = SpecConfig(**spec_config)

        # Set up NXentry and add misc. CHESS-specific metadata
        # as well as all spec_motors, scan_columns, and smb_pars
        nxentry = NXentry(name=spec_config.experiment_type)
        nxentry.spec_config = dumps(spec_config.dict())
        nxentry.attrs['station'] = spec_config.station
        nxentry.spec_scans = NXcollection()
        for scans in spec_config.spec_scans:
            nxscans = NXcollection()
            nxentry.spec_scans[f'{scans.scanparsers[0].scan_name}'] = nxscans
            nxscans.attrs['spec_file'] = str(scans.spec_file)
            nxscans.attrs['scan_numbers'] = scans.scan_numbers
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                nxscans[scan_number] = NXcollection()
                if hasattr(scanparser, 'spec_positioner_values'):
                    nxscans[scan_number].spec_motors = dumps(
                        {k:float(v) for k,v
                         in scanparser.spec_positioner_values.items()})
                if hasattr(scanparser, 'spec_scan_data'):
                    nxscans[scan_number].scan_columns = dumps(
                        {k:list(v) for k,v
                         in scanparser.spec_scan_data.items() if len(v)})
                if hasattr(scanparser, 'pars'):
                    nxscans[scan_number].smb_pars = dumps(
                        {k:v for k,v in scanparser.pars.items()})
                if detector_names:
                    nxdata = NXdata()
                    nxscans[scan_number].data = nxdata
                    for detector_name in detector_names:
                        nxdata[detector_name] = NXfield(
                           value=scanparser.get_detector_data(detector_name))

        return nxentry


class MapReader(Reader):
    """Reader for CHESS sample maps"""
    def read(self, filename=None, map_config=None, detector_names=[]):
        """Take a map configuration dictionary and return a
        representation of the map as an NXentry. The NXentry's default
        data group will contain the raw data collected over the course
        of the map.

        :param filename: name of file with the map configuration to
            read and pass onto the constructor of
            `CHAP.common.models.map.MapConfig`
        :type filename: str, optional
        :param map_config: map configuration to be passed directly to
            the constructor of `CHAP.common.models.map.MapConfig`
        :type map_config: dict, optional
        :param detector_names: Detector prefixes to include raw data
            for in the returned NXentry
        :type detector_names: list[str]
        :return: Data from the map configuration provided
        :rtype: nexusformat.nexus.NXentry
        """
        from json import dumps
        from nexusformat.nexus import (NXcollection,
                                       NXdata,
                                       NXentry,
                                       NXfield,
                                       NXsample)
        import numpy as np
        from CHAP.common.models.map import MapConfig

        if filename is not None:
            if map_config is not None:
                raise RuntimeError('Specify either filename or map_config '
                                   'in common.MapReader, not both')
            # Read the map configuration from file
            if not isfile(filename):
                raise OSError(f'input file does not exist ({filename})')
            extension = splitext(filename)[1]
            if extension in ('.yml', '.yaml'):
                reader = YAMLReader()
            else:
                raise RuntimeError('input file has a non-implemented '
                                   f'extension ({filename})')
            map_config = reader.read(filename)
        elif not isinstance(map_config, dict):
            raise RuntimeError('Invalid parameter map_config in '
                               f'common.MapReader ({map_config})')

        # Validate the map configuration provided by constructing a
        # MapConfig
        map_config = MapConfig(**map_config)

        # Set up NXentry and add misc. CHESS-specific metadata
        nxentry = NXentry(name=map_config.title)
        nxentry.attrs['station'] = map_config.station
        nxentry.map_config = dumps(map_config.dict())
        nxentry.spec_scans = NXcollection()
        for scans in map_config.spec_scans:
            nxentry.spec_scans[scans.scanparsers[0].scan_name] = \
                NXfield(value=scans.scan_numbers,
                        attrs={'spec_file': str(scans.spec_file)})

        # Add sample metadata
        nxentry[map_config.sample.name] = NXsample(**map_config.sample.dict())

        # Set up default data group
        nxentry.data = NXdata()
        nxentry.data.attrs['axes'] = map_config.dims
        for i, dim in enumerate(map_config.independent_dimensions[::-1]):
            nxentry.data[dim.label] = NXfield(
                value=map_config.coords[dim.label],
                units=dim.units,
                attrs={'long_name': f'{dim.label} ({dim.units})',
                       'data_type': dim.data_type,
                       'local_name': dim.name})
            nxentry.data.attrs[f'{dim.label}_indices'] = i

        # Create empty NXfields for all scalar data present in the
        # provided map configuration
        signal = False
        auxilliary_signals = []
        for data in map_config.all_scalar_data:
            nxentry.data[data.label] = NXfield(
                value=np.zeros(map_config.shape),
                units=data.units,
                attrs={'long_name': f'{data.label} ({data.units})',
                       'data_type': data.data_type,
                       'local_name': data.name})
            if not signal:
                signal = data.label
            else:
                auxilliary_signals.append(data.label)
        if signal:
            nxentry.data.attrs['signal'] = signal
            nxentry.data.attrs['auxilliary_signals'] = auxilliary_signals

        # Create empty NXfields of appropriate shape for raw
        # detector data
        for detector_name in detector_names:
            detector_data = map_config.get_detector_data(
                detector_name, (0,) * len(map_config.shape))
            nxentry.data[detector_name] = NXfield(value=np.zeros(
                (*map_config.shape, *detector_data.shape)),
                dtype=detector_data.dtype)
#            data_shape = list(map_config.shape)+list(detector_data.shape)
#            nxentry.data[detector_name] = NXfield(
#                value=np.zeros(data_shape), shape=data_shape,
#                dtype=detector_data.dtype)

        # Read and fill in maps of raw data
        if len(map_config.all_scalar_data) > 0 or len(detector_names) > 0:
            for map_index in np.ndindex(map_config.shape):
                for data in map_config.all_scalar_data:
                    nxentry.data[data.label][map_index] = map_config.get_value(
                        data, map_index)
                for detector_name in detector_names:
                    nxentry.data[detector_name][map_index] = \
                        map_config.get_detector_data(detector_name, map_index)

        return nxentry


class NexusReader(Reader):
    """Reader for NeXus files"""
    def read(self, filename, nxpath='/'):
        """Return the NeXus object stored at `nxpath` in the nexus
        file `filename`.

        :param filename: name of the NeXus file to read from
        :type filename: str
        :param nxpath: path to a specific loaction in the NeXus file
            to read from, defaults to `'/'`
        :type nxpath: str, optional
        :raises nexusformat.nexus.NeXusError: if `filename` is not a
            NeXus file or `nxpath` is not in `filename`.
        :return: the NeXus structure indicated by `filename` and `nxpath`.
        :rtype: nexusformat.nexus.NXobject
        """

        from nexusformat.nexus import nxload

        nxobject = nxload(filename)[nxpath]
        return nxobject


class URLReader(Reader):
    """Reader for data available over HTTPS"""
    def read(self, url, headers={}, timeout=10):
        """Make an HTTPS request to the provided URL and return the
        results.  Headers for the request are optional.

        :param url: the URL to read
        :type url: str
        :param headers: headers to attach to the request, defaults to
            `{}`
        :type headers: dict, optional
        :return: the content of the response
        :rtype: object
        """

        import requests

        resp = requests.get(url, headers=headers, timeout=timeout)
        data = resp.content

        self.logger.debug(f'Response content: {data}')

        return data


class YAMLReader(Reader):
    """Reader for YAML files"""
    def read(self, filename):
        """Return a dictionary from the contents of a yaml file.

        :param filename: name of the YAML file to read from
        :return: the contents of `filename`
        :rtype: dict
        """

        import yaml

        with open(filename) as file:
            data = yaml.safe_load(file)
        return data


if __name__ == '__main__':
    from CHAP.reader import main
    main()



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#pylint: disable=
"""
File       : server.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Python server with thread pool and CHAP pipeline

### Client side:
cat /tmp/chap.json
{
  "pipeline": [{"common.PrintProcessor": {}}],
  "input": 1
}

### curl call to the server with our CHAP pipeline
curl -X POST -H "Content-type: application/json" -d@/tmp/chap.json http://localhost:5000/pipeline
{"pipeline":[{"common.PrintProcessor":{}}],"status":"ok"}

### Server side:
flask --app server run
 * Serving Flask app 'server'
 * Debug mode: off
WARNING: This is a development server. Do not use it in a production deployment. Use a production WSGI server instead.
 * Running on http://127.0.0.1:5000
Press CTRL+C to quit
...

CHAP.server         : call pipeline args=() kwds={'pipeline': [{'common.PrintProcessor': {}}]}
CHAP.server         : pipeline
[{'common.PrintProcessor': {}}]
CHAP.server         : Loaded <CHAP.common.processor.PrintProcessor object at 0x10e0f1ed0>
CHAP.server         : Loaded <CHAP.pipeline.Pipeline object at 0x10e0f1f10> with 1 items

CHAP.server         : Calling "execute" on <CHAP.pipeline.Pipeline object at 0x10e0f1f10>
Pipeline            : Executing "execute"

Pipeline            : Calling "process" on <CHAP.common.processor.PrintProcessor object at 0x10e0f1ed0>
PrintProcessor      : Executing "process" with type(data)=<class 'NoneType'>
PrintProcessor data :
None
PrintProcessor      : Finished "process" in 0.000 seconds

Pipeline            : Executed "execute" in 0.000 seconds
127.0.0.1 - - [07/Apr/2023 09:11:22] "POST /pipeline HTTP/1.1" 200 -
"""

# system modules
import time
import logging
from queue import Queue

# thrid-party modules

# Flask modules
from flask import Flask, request, jsonify

# CHAP modules
from CHAP.TaskManager import TaskManager, start_new_thread 
from CHAP.runner import run, setLogger


# Task manager to execute our tasks
taskManager = TaskManager()

# Flask Server
app = Flask(__name__)

# daemon task queue
task_queue = Queue()

@app.route("/")
def index_route():
    """
    Server main end-point
    """
    return "CHAP daemon"

@app.route("/run")
def run_route():
    """
    Server main end-point
    """
    task = request.args.get('task')
    task_queue.put(task)
    return f"Execute {task}"

@app.route("/pipeline", methods=["POST"])
def pipeline_route():
    """
    Server /pipeline end-point
    """
    content = request.json
    if 'pipeline' in content:
        # spawn new pipeline task
        jobs = []
        jobs.append(taskManager.spawn(task, pipeline=content['pipeline']))
        taskManager.joinall(jobs)
        return {"status": "ok", "pipeline": content['pipeline']}
    else:
        return {"status": "fail", "reason": "no pipeline in incoming request"}

def task(*args, **kwds):
    """
    Helper function to execute CHAP pipeline
    """
    log_level = "INFO"
    logger, log_handler = setLogger(log_level)
    logger.info(f"call pipeline args={args} kwds={kwds}")
    pipeline = kwds['pipeline']
    logger.info(f"pipeline\n{pipeline}")
    run(pipeline, logger, log_level, log_handler)

def daemon(name, queue, interval):
    """
    Daemon example based on Queue
    """
    print(f"Daemon {name}")
    while True:
        if queue.qsize() == 0:
            print("Default action")
            time.sleep(interval)
        else:
            task = queue.get()
            if task == "exit":
                return
            print(f"daemon run {task}")

# start daemon thread in addition to Flask server
start_new_thread("daemon", daemon, ("daemon", task_queue, 3))



#!/usr/bin/env python
"""
File       : writer.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: generic Writer module

Define a generic `Writer` object.
"""

# system modules
import argparse
from inspect import getfullargspec
import logging
from sys import modules
from time import time

# local modules
from CHAP.pipeline import PipelineItem


class Writer(PipelineItem):
    """Generic file writer

    The job of any `Writer` in a `Pipeline` is to receive input
    returned by a previous `PipelineItem`, write its data to a
    particular file format, then return the same data unaltered so it
    can be used by a successive `PipelineItem`.
    """

    def write(self, data, filename):
        """Write the input data as text to a file.

        :param data: input data
        :type data: list[PipelineData]
        :param filename: Name of file to write to
        :type filename: str
        :return: contents of the input data
        :rtype: object
        """

        data = self.unwrap_pipelinedata(data)[-1]
        with open(filename, 'a') as file:
            file.write(data)
        return data


class OptionParser():
    """User based option parser"""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--data', action='store',
            dest='data', default='', help='Input data')
        self.parser.add_argument(
            '--filename', action='store',
            dest='filename', default='', help='Output file')
        self.parser.add_argument(
            '--writer', action='store',
            dest='writer', default='Writer', help='Writer class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function"""

    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.writer
    try:
        writer_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported writer {cls_name}')
        raise

    writer = writer_cls()
    writer.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    writer.logger.addHandler(log_handler)
    data = writer.write(opts.data, opts.filename)
    print(f'Writer {writer} writes to {opts.filename}, data {data}')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Standard python setup.py file
to build     : python setup.py build
to install   : python setup.py install --prefix=<some dir>
to clean     : python setup.py clean
to build doc : python setup.py doc
to run tests : python setup.py test
"""

import os
import setuptools

# [set version]
version = 'v0.0.11'
# [version set]

def datafiles(idir, pattern=None):
    """Return list of data files in provided relative dir"""
    files = []
    for dirname, dirnames, filenames in os.walk(idir):
        for subdirname in dirnames:
            files.append(os.path.join(dirname, subdirname))
        for filename in filenames:
            if  filename[-1] == '~':
                continue
            # match file name pattern (e.g. *.css) if one given
            if pattern and not fnmatch.fnmatch(filename, pattern):
                continue
            files.append(os.path.join(dirname, filename))
    return files

data_files = datafiles('examples')

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ChessAnalysisPipeline",
    version=version,
    author="Keara Soloway, Rolf Verberg, Valentin Kuznetsov",
    author_email="",
    description="CHESS analysis pipeline framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CHESSComputing/ChessAnalysisPipeline",
    packages=[
        'CHAP',
        'CHAP.common',
        'CHAP.common.models',
        'CHAP.edd',
        'CHAP.inference',
        'CHAP.saxswaxs',
        'CHAP.sin2psi',
        'CHAP.tomo',
        'CHAP.utils',
        'MLaaS'
    ],
    package_dir={
        'CHAP': 'CHAP',
        'CHAP.common': 'CHAP/common',
        'CHAP.common.models': 'CHAP/common/models',
        'CHAP.edd': 'CHAP/edd',
        'CHAP.inference': 'CHAP/inference',
        'CHAP.saxswaxs': 'CHAP/saxswaxs',
        'CHAP.sin2psi': 'CHAP/sin2psi',
        'CHAP.tomo': 'CHAP/tomo',
        'CHAP.utils': 'CHAP/utils',
        'MLaaS': 'MLaaS'
    },
    package_data={
        'examples': data_files
    },
    entry_points={
        'console_scripts': ['CHAP = CHAP.runner:main']
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'PyYAML'
    ],
)
